#!/usr/bin/env python3

import isisdl.compress as compress
from isisdl.backend import sync_database
from isisdl.backend.config import init_wizard, config_wizard
from isisdl.backend.request_helper import CourseDownloader
from isisdl.settings import is_first_time, is_static, current_database_version, forbidden_chars, has_ffmpeg, fstype
from isisdl.settings import is_online
from isisdl.utils import args, acquire_file_lock_or_exit, generate_error_message, install_latest_version, export_config, database_helper, \
    config, migrate_database
from isisdl.version import __version__


def _main() -> None:
    if is_first_time:
        print("""
It seems as if this is your first time executing isisdl. Welcome 💖

I will guide you through a short configuration phase of about 5min.
It is recommended that you read these options carefully.

If you wish to re-configure me, run `isisdl --init` or `isisdl --config`.


Please press enter to continue.
""")
        input()
        init_wizard()
        config_wizard()
        sync_database.main()

    elif database_helper.get_database_version() < config.default("database_version"):
        if migrate_database() is False:
            exit(1)

        exit(0)

    elif args.version:
        print(f"""isisdl version {__version__}

Build info:

{is_static = }
{current_database_version = }
{has_ffmpeg = }
{forbidden_chars = }
{fstype = }
""")
        exit(0)

    acquire_file_lock_or_exit()

    if args.init:
        init_wizard()
        exit(0)

    elif args.config:
        config_wizard()
        exit(0)

    elif args.export_config:
        print("Exporting current configuration ...")
        export_config()
        exit(0)

    elif args.delete_bad_urls:
        print("Deleting bad urls ...")
        database_helper.delete_bad_urls()
        exit(0)

    if not is_online:
        print("I cannot establish an internet connection.")
        exit(1)

    install_latest_version()

    if args.update:
        print("No new update available. (Cricket sounds ...)")
        exit(0)

    elif args.sync:
        sync_database.main()
        exit(0)

    elif args.compress:
        compress.main()
        exit(0)

    elif args.subscribe:
        print("Due to legal reasons, this is currently not supported. :(")
        exit(0)

#         print("""Attention:
# This option will lead to you subscribing to *every* publicly available ISIS course.
#
# Subscribing will be quite fast 10-20s, but unsubscribing takes a few minutes.
# This behaviour is due to the fact that the API to unsubscribe from courses
# is not yet implemented. (https://tracker.moodle.org/browse/MDL-64255)
#
# Please press enter to continue.
# """)
#
#         input()
#         subscribe_to_all_courses()
#         exit(0)

    elif args.unsubscribe:
        print("Due to legal reasons, this is currently not supported. :(")
        exit(0)

        # unsubscribe_from_courses()
        # exit(0)

    else:
        # Main routine
        CourseDownloader().start()
        CourseDownloader.shutdown_running_downloads()
        print("Done! Have a nice day ^.^")


def main() -> None:
    try:
        _main()
    except Exception as ex:
        generate_error_message(ex)


# Feature discussion:
#   Windows autorun
#   Download of corrupted files
#   Streaming files: Is it worth it?


# Suggestion for ISIS API:
#   video server also sends the size

# Main TODOS:
#   Dynamic calc
#   compress


# No size implies:
#   Harder time conflicts in files
#   No download status file size
#   No sync database

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import atexit
import enum
import json
import os
import platform
import random
import re
import shutil
import signal
import stat
import string
import subprocess
import sys
import time
import traceback
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from functools import wraps
from itertools import repeat
from pathlib import Path
from queue import PriorityQueue, Queue, Full, Empty
from tempfile import TemporaryDirectory
from threading import Thread
from typing import Callable, List, Tuple, Dict, Any, Set, cast, Iterable, NoReturn, TYPE_CHECKING, DefaultDict
from typing import Optional, Union
from urllib.parse import unquote, parse_qs, urlparse

import colorama
import distro as distro
import requests
from packaging import version
from packaging.version import Version, LegacyVersion
from requests import Session

from isisdl import settings
from isisdl.backend.database_helper import DatabaseHelper
from isisdl.settings import download_chunk_size, token_queue_download_refresh_rate, forbidden_chars, replace_dot_at_end_of_dir_name, force_filesystem
from isisdl.settings import working_dir_location, is_windows, checksum_algorithm, checksum_num_bytes, example_config_file_location, config_dir_location, database_file_location, status_time, \
    discover_num_threads, status_progress_bar_resolution, download_progress_bar_resolution, config_file_location, is_first_time, is_autorun, parse_config_file, lock_file_location, \
    enable_lock, error_directory_location, systemd_dir_location, master_password, is_testing, systemd_timer_file_location, systemd_service_file_location, export_config_file_location, \
    isisdl_executable, is_static, enable_multithread, subscribe_num_threads, subscribed_courses_file_location, error_text, token_queue_refresh_rate
from isisdl.version import __version__

if TYPE_CHECKING:
    from isisdl.backend.status import Status
    from isisdl.backend.request_helper import RequestHelper


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="isisdl", formatter_class=argparse.RawTextHelpFormatter, description="""
    This program downloads all content from your ISIS profile.""")

    parser.add_argument("-t", "--max-num-threads", help="The maximum number of threads to spawn (for downloading files)\n ", type=int, default=3, metavar="{num}")
    parser.add_argument("-d", "--download-rate", help="Limits the download rate to {num} MiB/s\n ", type=float, default=None, metavar="{num}")

    operations = parser.add_mutually_exclusive_group()

    operations.add_argument("-v", "--version", help="Print the version number and exit", action="store_true")
    operations.add_argument("--init", help="Guides you through the initial configuration and setup process.", action="store_true")
    operations.add_argument("--config", help="Guides you through additional configuration which focuses on what to download from ISIS.", action="store_true")
    operations.add_argument("--sync", help="Synchronizes the local database with ISIS. Will delete not existent or corrupted entries.", action="store_true")
    operations.add_argument("--compress", help="Starts ffmpeg and will compress all downloaded videos.", action="store_true")
    operations.add_argument("--subscribe", help="Subscribes you to *all* ISIS courses publicly available.", action="store_true")
    operations.add_argument("--unsubscribe", help="Unsubscribes you from the courses you got subscribed by running `isisdl --subscribe`.", action="store_true")
    operations.add_argument("--export-config", help=f"Exports the config to {export_config_file_location}", action="store_true")
    operations.add_argument("--stream", help="Launches isisdl in streaming mode. Will watch for file accesses and download only those files.", action="store_true")
    operations.add_argument("--update", help="Checks if an update is available and installs it.", action="store_true")
    operations.add_argument("--delete-bad-urls", help="Deletes all urls deemed to be \"bad\" - meaning there is no content.", action="store_true")

    if is_testing:
        return parser.parse_known_args()[0]

    try:
        return parser.parse_args()
    except SystemExit as ex:
        os._exit(ex.code)


class Config:
    password_encrypted: Optional[bool]
    username: Optional[str]
    password: Optional[str]

    whitelist: Optional[List[int]]
    blacklist: Optional[List[int]]
    renamed_courses: Optional[Dict[int, str]]
    make_subdirs: bool
    follow_links: bool

    download_videos: bool
    filename_replacing: bool
    timer_enable: bool
    throttle_rate: Optional[int]
    throttle_rate_autorun: Optional[int]
    update_policy: Optional[str]
    telemetry_policy: bool
    database_version: int
    absolute_path_filename: bool

    auto_subscribed_courses: Optional[List[int]]

    __slots__ = tuple(__annotations__)  # type: ignore

    default_config: Dict[str, Union[bool, str, int, None, Dict[int, str]]] = {
        "password_encrypted": False,
        "username": None,
        "password": None,

        "whitelist": None,
        "blacklist": None,
        "renamed_courses": None,
        "make_subdirs": True,
        "follow_links": True,

        "download_videos": True,
        "filename_replacing": False,
        "timer_enable": True,
        "throttle_rate": None,
        "throttle_rate_autorun": None,
        "update_policy": "install_pip",
        "telemetry_policy": True,
        "absolute_path_filename": False,
        "database_version": 2,

        "auto_subscribed_courses": None,
    }

    state: Dict[str, Union[bool, str, int, None, Dict[int, str]]] = {k: None for k in default_config}  # The state to consider after defaults, config files etc.
    _stored: Dict[str, Union[bool, str, int, None, Dict[int, str]]] = {}  # Values the user has actively stored in the config wizard (no config file)
    _backup: Dict[str, Union[bool, str, int, None, Dict[int, str]]] = {}  # Extra backup to maintain for tests
    _in_backup: bool = False

    def __init__(self, _prev_config: Optional[Dict[str, Any]] = None) -> None:
        config_file_data = parse_config_file()
        stored_config = database_helper.get_config()
        prev_config = _prev_config or {}

        # Filter out keys for settings
        for k, v in list(config_file_data.items()):
            if k in settings.__dict__:
                del config_file_data[k]

        # Verify config keys
        global_vars, default_keys = settings.global_vars, Config.default_config.keys()
        for k in list(config_file_data.keys()) + list(stored_config.keys()):
            if k not in global_vars and k not in default_keys:
                print(f"{error_text} the config file is malformed.")
                print(f"The file is located at `{config_file_location}`\n\n")
                print(f"Reason: Unrecognized key: {repr(k)}\n")
                os._exit(1)

        # Json only allows for keys to be strings (https://stackoverflow.com/a/8758771)
        # Set the keys manually back to ints, so we can work with them.
        for item in [config_file_data, stored_config]:
            if "renamed_courses" in item and item["renamed_courses"] is not None:
                assert isinstance(item["renamed_courses"], dict)
                item["renamed_courses"] = {int(k): v for k, v in item["renamed_courses"].items()}

        assert all(k in self.__slots__ for k in prev_config)

        self.state.update(Config.default_config)
        self.state.update(stored_config)
        self.state.update(prev_config)
        self.state.update(config_file_data)

        self._stored.update(stored_config)
        self.state["database_version"] = database_helper.get_database_version()

        if "database_version" not in self._stored:
            self._stored["database_version"] = self.default("database_version")

        if database_helper.get_database_version() == Config.default("database_version"):
            self.verify_state_types()

        for name in self.__slots__:
            super().__setattr__(name, Config.state[name])

    def __setattr__(self, key: str, value: Union[bool, str, int, None, Dict[int, str]]) -> None:
        super().__setattr__(key, value)
        if not self._in_backup:
            Config.state[key] = value
            Config._stored[key] = value
            database_helper.set_config(self._stored)

    def verify_state_types(self) -> None:
        # TODO: Use Config.__annotations__
        # Unfortunately we can't use the type annotations since they provide very little interface to python.
        def fail(attr: str, typ: Any, may_be_none: bool = False) -> None:
            if self.state[attr] is None and may_be_none:
                return

            if type(self.state[attr]) is not typ:
                print(f"{error_text} the config file is malformed.\n"
                      f"Reason: Expected type {typ} for key {repr(attr)}. Got {type(self.state[attr])}.\nBailing out!")
                os._exit(1)

        fail("password_encrypted", bool, True)
        fail("username", str, True)
        fail("password", str, True)
        fail("whitelist", list, True)
        fail("blacklist", list, True)
        fail("renamed_courses", dict, True)
        fail("make_subdirs", bool)
        fail("follow_links", bool)
        fail("download_videos", bool)
        fail("timer_enable", bool)
        fail("throttle_rate", int, True)
        fail("throttle_rate_autorun", int, True)
        fail("update_policy", str, True)
        fail("telemetry_policy", bool)
        fail("database_version", int)
        fail("absolute_path_filename", bool)

    @staticmethod
    def default(attr: str) -> Any:
        return Config.default_config[attr]

    @staticmethod
    def user(attr: str) -> Optional[Any]:
        return Config._stored.get(attr, None)

    def to_dict(self) -> Dict[str, Union[bool, str, int, None, Dict[int, str]]]:
        return {name: getattr(self, name) for name in self.__slots__}

    def start_backup(self) -> None:
        Config._backup = self.to_dict()
        Config._in_backup = True

    def restore_backup(self) -> None:
        Config._in_backup = False
        for name in self.__slots__:
            super().__setattr__(name, self._backup[name])


def encode_yaml(st: Union[bool, str, int, None, Dict[int, str]]) -> str:
    if st is None:
        return "null"
    elif st is True:
        return "true"
    elif st is False:
        return "false"
    return str(st)


def generate_config_str(
        working_dir_location: str, database_file_location: str, master_password: str, filename_replacing: bool, download_videos: bool, whitelist: Optional[List[int]], blacklist: Optional[List[int]],
        throttle_rate: Optional[int], throttle_rate_autorun: Optional[int], update_policy: Optional[str], telemetry_policy: bool, status_time: float, video_size_discover_num_threads: int,
        status_progress_bar_resolution: int, download_progress_bar_resolution: int, force_filesystem: Optional[str], make_subdirs: bool, follow_links: bool, absolute_path_filename: bool
) -> str:
    return f"""---

# Any values you overwrite here will take precedence over *any* otherwise provided value.


# The directory where everything lives in.
# Possible values {{any absolute posix path}}
working_dir_location: {working_dir_location}

# The name of the SQlite Database (located in `working_dir_location`) used for storing metadata about files + config.
# Possible values {{any posix path}}
database_file_location: {database_file_location}

# The password to encrypt your password, if none is provided
# Possible values {{any string}}
master_password: {master_password}


# The time waited between re-renders of the status message.
# If you have a fast terminal / PC you can easily set this value to 0.1 or even 0.01.
# Possible values {{any float}}
status_time: {status_time}


# Should videos be downloaded on this device?
# Possible values {{"true", "false"}}
download_videos: {encode_yaml(download_videos)}


# Should the filename be replaced with a sanitized version?
# Possible values {{"true", "false"}}
filename_replacing: {encode_yaml(filename_replacing)}


# The global whitelist of courses to be considered. Best set with `isisdl-config` and then extracted.
# Possible values {{"null", list[int] of course ID's}}
whitelist: {'null' if whitelist is None else whitelist}

# The global blacklist of courses to be considered. Best set with `isisdl-config` and then extracted.
# Possible values {{"null", list[int] of course ID's}}
blacklist: {'null' if blacklist is None else blacklist}


# Should subdirectories be created?
# Possible values {{"true", "false"}}
make_subdirs: {encode_yaml(make_subdirs)}


# Should external links be followed?
# Possible values {{"true", "false"}}
follow_links: {encode_yaml(follow_links)}


# Should the displayed name be absolute or just the name?
# Possible values {{"true", "false"}}
absolute_path_filename: {encode_yaml(absolute_path_filename)}


# The global throttle rate. Will take precedence over throttle_rate_autorun.
# Possible values {{"null", any integer}}
throttle_rate: {encode_yaml(throttle_rate)}

# The throttle rate for when `isisdl` automatically runs.
# Possible values {{"null", any integer}}
throttle_rate_autorun: {encode_yaml(throttle_rate_autorun)}


# How updates should be handled.
# Possible values {{"null", "install_pip", "install_github", "notify_pip", "notify_github"}}
update_policy: {encode_yaml(update_policy)}


# Should telemetry data be collected?
# Possible values {{"true", "false"}}
telemetry_policy: {encode_yaml(telemetry_policy)}


# Force a filesystem
# Possible values {{null, ext, ext2, ext3, ext4, btrfs, exfat, fat32, vfat, ntfs, hfs, hfsplus}}
force_filesystem: {force_filesystem}


# Number of threads to use when discovering video file sizes (ISIS does not offer an API).
# Possible values {{any integer > 0}}
video_size_discover_num_threads: {video_size_discover_num_threads}


# The resolution of any ordinary status has.
# Possible values {{any integer > 0}}
status_progress_bar_resolution: {status_progress_bar_resolution}

# The resolution of the download progress bar.
# Possible values {{any integer > 0}}
download_progress_bar_resolution: {download_progress_bar_resolution}
"""


def generate_default_config_str() -> str:
    return generate_config_str(
        working_dir_location, database_file_location, master_password, Config.default("filename_replacing"), Config.default("download_videos"), Config.default("whitelist"),
        Config.default("blacklist"), Config.default("throttle_rate"), Config.default("throttle_rate_autorun"), Config.default("update_policy"), Config.default("telemetry_policy"), status_time,
        discover_num_threads, status_progress_bar_resolution, download_progress_bar_resolution, force_filesystem, Config.default("make_subdirs"), Config.default("follow_links"),
        Config.default("absolute_path_filename")
    )


def generate_current_config_str() -> str:
    return generate_config_str(
        working_dir_location, database_file_location, master_password, config.filename_replacing, config.download_videos, config.whitelist, config.blacklist, config.throttle_rate,
        config.throttle_rate_autorun, config.update_policy, config.telemetry_policy, status_time, discover_num_threads, status_progress_bar_resolution, download_progress_bar_resolution,
        force_filesystem, config.make_subdirs, config.follow_links, config.absolute_path_filename
    )


def export_config() -> None:
    with open(export_config_file_location, "w") as f:
        f.write(generate_current_config_str())


def startup() -> None:
    os.makedirs(path(), exist_ok=True)
    if os.path.exists(path(error_directory_location)) and os.listdir(path(error_directory_location)) == []:
        os.rmdir(path(error_directory_location))

    if not is_windows:
        os.makedirs(path(config_dir_location), exist_ok=True)
        os.makedirs(systemd_dir_location, exist_ok=True)

        default_config_str = generate_default_config_str()

        with open(example_config_file_location, "w") as f:
            f.write(default_config_str)

        if not os.path.exists(config_file_location):
            with open(config_file_location, "w") as f:
                f.write(f"# You probably want to start by copying {example_config_file_location} and adapting it.\n")


def clear() -> None:
    if is_windows:
        os.system('cls')
    else:
        os.system('clear')


def migrate_database() -> bool:
    print(f"""
I have detected a breaking change in the database.
In order to account for these changes, I will have to delete the database,
make a few changes to local files and rediscover all the files.

If something goes wrong, simply delete the directory
`{path()}`
and everything will get downloaded as usual.

Please confirm that this is okay. [y/n]""")
    choice = get_input({"y", "n"})
    if choice == "n":
        print("Alright, I am not doing any changes.")
        return False

    from isisdl.backend.crypt import get_credentials
    from isisdl.backend.request_helper import RequestHelper
    global config
    helper = RequestHelper(get_credentials(["peanuts"]))

    def migrate_1_to_2() -> None:
        course_name_counts: DefaultDict[str, int] = defaultdict(int)
        for course in helper._courses:
            course_name_counts[course.old_name] += 1

        for _course, num in course_name_counts.items():
            if num > 1:
                try:
                    shutil.rmtree(path(_course))
                except OSError:
                    pass

        downloaded_courses = set(os.listdir(path()))
        for course in helper._courses:
            if course.name == course.old_name:
                continue

            if course.old_name in downloaded_courses:
                if os.path.exists(path(course.name)):
                    shutil.rmtree(path(course.name))

                try:
                    os.rename(path(course.old_name), path(course.name))
                except OSError as ex:
                    generate_error_message(ex)

        config.database_version = 2

    while database_helper.get_database_version() < config.default("database_version"):
        eval(f"migrate_{database_helper.get_database_version()}_to_{database_helper.get_database_version() + 1}()")

    os.unlink(path(database_file_location))

    print("\nSuccessfully migrated.\nPlease restart me and I will guide you through the new configuration!")

    return True


def parse_google_drive_url(url: str) -> Optional[str]:
    """
    Copied from https://github.com/wkentaro/gdown

    Parse URLs especially for Google Drive links.
    drive_id: ID of file on Google Drive.
    is_download_link: Flag if it is download link of Google Drive.
    """
    parsed = urlparse(url)
    query = parse_qs(parsed.query)

    drive_id = None
    if "id" in query:
        drive_ids = query["id"]
        if len(drive_ids) == 1:
            drive_id = drive_ids[0]
    else:
        patterns = [r"^/file/d/(.*?)/view$", r"^/presentation/d/(.*?)/edit$"]
        for pattern in patterns:
            match = re.match(pattern, parsed.path)
            if match:
                drive_id = match.groups()[0]
                break

    return drive_id


def get_url_from_gdrive_confirmation(contents: str) -> Optional[str]:
    """
    Copied from https://github.com/wkentaro/gdown
    """

    url = ""
    for line in contents.splitlines():
        m = re.search(r'href="(\/uc\?export=download[^"]+)', line)
        if m:
            url = "https://docs.google.com" + m.groups()[0]
            url = url.replace("&amp;", "&")
            break
        m = re.search('id="downloadForm" action="(.+?)"', line)
        if m:
            url = m.groups()[0]
            url = url.replace("&amp;", "&")
            break
        m = re.search('"downloadUrl":"([^"]+)', line)
        if m:
            url = m.groups()[0]
            url = url.replace("\\u003d", "=")
            url = url.replace("\\u0026", "&")
            break
        m = re.search('<p class="uc-error-subcaption">(.*)</p>', line)
        if m:
            return None

    if not url:
        return None

    return url


def run_cmd_with_error(args: List[str]) -> None:
    result = subprocess.run(args, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

    if result.returncode:
        print(error_text)
        print(f"The command `{' '.join(result.args)}` exited with exit code {result.returncode}\n{result.stdout.decode()}{result.stderr.decode()}")
        print("\nPress enter to continue")
        input()


def do_ffprobe(file: Path) -> Optional[Dict[str, Any]]:
    # This function is copied and adapted from ffmpeg-python: https://github.com/kkroening/ffmpeg-python
    args = ["ffprobe", "-show_format", "-show_streams", "-of", "json", "-show_data_hash", "sha256", "-i", str(file)]

    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    if p.returncode != 0:
        return None

    return cast(Dict[str, Any], json.loads(out.decode('utf-8')))


def is_h265(file: Path) -> Optional[bool]:
    probe = do_ffprobe(file)
    if probe is None:
        return None

    video_stream = next((stream for stream in probe['streams'] if stream['codec_type'] == 'video'), None)
    if video_stream is None:
        return None

    if "codec_name" not in video_stream:
        logger.assert_fail('"codec_name" not in video_stream')
        return None

    return bool(video_stream["codec_name"] == "hevc")


def check_pypi_for_version() -> Optional[Union[LegacyVersion, Version]]:
    # Inspired from https://pypi.org/project/pypi-search
    to_search = requests.get("https://pypi.org/project/isisdl/").text
    found_version = re.search("<h1 class=\"package-header__name\">\n *(.*)?\n *</h1>", to_search)

    if found_version is None:
        return None

    groups = found_version.groups()
    if groups is None or len(groups) != 1:
        return None

    return version.parse(groups[0].split()[1])


def check_github_for_version() -> Optional[Union[LegacyVersion, Version]]:
    if is_static:
        latest_release = requests.get("https://api.github.com/repos/Emily3403/isisdl/releases/latest").json()
        if "tag_name" not in latest_release:
            return None

        return version.parse(latest_release["tag_name"][1:])

    else:
        badge = requests.get("https://github.com/Emily3403/isisdl/actions/workflows/tests.yml/badge.svg").text
        if "passing" not in badge:
            return None

        res = requests.get("https://raw.githubusercontent.com/Emily3403/isisdl/main/src/isisdl/version.py")
        if not res.ok:
            return None

        found_version = re.match("__version__ = \"(.*)?\"", res.text)
        if found_version is None:
            return None

        return version.parse(found_version.group(1))


def install_latest_version() -> None:
    if is_first_time:
        return

    if config.update_policy is None:
        return

    version_github = check_github_for_version()
    version_pypi = check_pypi_for_version()

    new_version = version_github if config.update_policy.endswith("github") else version_pypi

    if new_version is None:
        return

    if new_version <= version.parse(__version__):
        return

    print(f"\nThere is a new version of isisdl available: {new_version} (current: {__version__}).")

    if config.update_policy.startswith("notify") and not args.update:
        return
    elif config.update_policy.startswith("install"):
        print("According to your update policy I will auto-install it.\n")

    if "pip" in config.update_policy:
        ret = subprocess.call([sys.executable, "-m", "pip", "install", "--upgrade", "isisdl"])

    elif "github" in config.update_policy and is_static is False:
        ret = subprocess.call([sys.executable, "-m", "pip", "install", "git+https://github.com/Emily3403/isisdl"])

    # TODO: test if this will work
    elif "github" in config.update_policy and is_static:
        with TemporaryDirectory() as tmp:
            assets = requests.get("https://api.github.com/repos/Emily3403/isisdl/releases/latest").json()["assets"]
            correct_name = "isisdl-windows.exe" if is_windows else "isisdl-linux.bin"
            for asset in assets:
                if asset["name"] == correct_name:
                    break

            else:
                print(f"{error_text} I cannot find the release for your platform.")
                exit(1)

            new_file = requests.get(asset["browser_download_url"], stream=True)
            new_isisdl = os.path.join(tmp, correct_name)
            with open(new_isisdl, "wb") as f:
                shutil.copyfileobj(new_file.raw, f)

            st = os.stat(new_isisdl)
            os.chmod(new_isisdl, st.st_mode | stat.S_IEXEC)
            os.replace(isisdl_executable, new_isisdl)
            ret = 0

    else:
        assert False

    if ret == 0:
        print("\n\nSuccessfully updated!")
        exit(0)
    else:
        print("\n\nUpdating failed… why?")
        exit(ret)


def path(*args: str) -> Path:
    return Path(working_dir_location, *args)


def remove_systemd_timer() -> None:
    if not os.path.exists(systemd_timer_file_location):
        return

    run_cmd_with_error(["systemctl", "--user", "disable", "--now", "isisdl.timer"])
    run_cmd_with_error(["systemctl", "--user", "daemon-reload"])

    if os.path.exists(systemd_timer_file_location):
        os.remove(systemd_timer_file_location)

    if os.path.exists(systemd_service_file_location):
        os.remove(systemd_service_file_location)


def install_systemd_timer() -> None:
    import isisdl.autorun
    with open(systemd_service_file_location, "w") as f:
        f.write(f"""# isisdl autorun service
# This file was autogenerated by `isisdl --init`.

[Unit]
Description=isisdl autorun
Wants=isisdl.timer

[Service]
Type=oneshot
ExecStart={isisdl_executable} {isisdl.autorun.__file__}

[Install]
WantedBy=multi-user.target
""")

    with open(systemd_timer_file_location, "w") as f:
        f.write("""# isisdl autorun timer
# This file was autogenerated by the `isisdl-config` utility.

[Unit]
Description=isisdl
Wants=isisdl.service

[Timer]
Unit=isisdl.service
OnCalendar=hourly

[Install]
WantedBy=timers.target
""")

    run_cmd_with_error(["systemctl", "--user", "enable", "isisdl.timer"])
    run_cmd_with_error(["systemctl", "--user", "daemon-reload"])


def sanitize_name(name: str, is_dir: bool) -> str:
    # Remove unnecessary whitespace
    name = name.strip()
    name = unquote(name)

    # First replace umlaute
    for a, b in {"ä": "a", "ö": "o", "ü": "u"}.items():
        name = name.replace(a, b)
        name = name.replace(a.upper(), b.upper())

    # Now start to add chars that don't fit.
    char_string = ""

    # Now replace any remaining funny symbols with a `?`
    name = name.encode("ascii", errors="replace").decode()

    if config.filename_replacing:
        char_string += r"""!"#$%&'()*+,/:;<=>?@[\]^`{|}~"""
        name = name.translate(str.maketrans(char_string, "\0" * len(char_string)))

    # This is probably a suboptimal solution, but it works…
    str_list = list(name)
    final = []

    if config.filename_replacing:
        whitespaces = set(string.whitespace + "_-")
        i = 0
        next_upper = False
        while i < len(str_list):
            char = str_list[i]

            if char == "\0":
                pass
            elif char in whitespaces:
                next_upper = True

            else:
                if next_upper:
                    final.append(char.upper())
                    next_upper = False
                else:
                    final.append(char)

            i += 1
    else:
        final = str_list

    # Folder names cannot end with a period in Windows ...
    while replace_dot_at_end_of_dir_name and is_dir and final and (final[-1] == "." or final[-1] == " "):
        final.pop()

    final_str = "".join(item for item in final if item not in forbidden_chars)

    # All supported filesystems have a limit of 255 bytes for the filename. Enforce it.
    return final_str.encode()[:255].decode()


def get_input(allowed: Set[str]) -> str:
    while True:
        choice = input()
        if choice in allowed:
            break

        print(f"Unhandled character: {choice!r} is not in the expected {{" + ", ".join(repr(item) for item in sorted(list(allowed))) + "}\nPlease try again.\n")

    return choice


class OnKill:
    _funcs: PriorityQueue[Tuple[int, Callable[[], None]]] = PriorityQueue()
    _min_priority = 0
    _already_killed = False
    _pids_to_kill: List[int] = []

    def __init__(self) -> None:
        signal.signal(signal.SIGINT, OnKill.exit)
        signal.signal(signal.SIGABRT, OnKill.exit)
        signal.signal(signal.SIGTERM, OnKill.exit)
        signal.signal(signal.SIGILL, OnKill.exit)
        signal.signal(signal.SIGSEGV, OnKill.exit)

        if sys.platform == "win32":
            signal.signal(signal.SIGBREAK, OnKill.exit)
        else:
            signal.signal(signal.SIGBUS, OnKill.exit)
            signal.signal(signal.SIGHUP, OnKill.exit)

    @staticmethod
    def add(func: Any, priority: Optional[int] = None) -> None:
        if priority is None:
            # Generate a new priority → max priority
            priority = OnKill._min_priority - 1

        OnKill._min_priority = min(priority, OnKill._min_priority)

        OnKill._funcs.put((priority, func))

    @staticmethod
    @atexit.register
    def exit(sig: Optional[int] = None, frame: Any = None) -> None:
        if sig is None:
            OnKill.do_funcs()
            return

        if OnKill._already_killed:
            print("Alright, stay calm. I am skipping cleanup and exiting!")

            # Kill remaining processes
            for pid in OnKill._pids_to_kill:
                try:
                    if sys.platform == "win32":
                        os.kill(pid, signal.SIGABRT)
                    else:
                        os.kill(pid, signal.SIGKILL)
                except Exception:
                    pass

            os._exit(sig)

        OnKill._already_killed = True
        OnKill.do_funcs()
        os._exit(0)

    @staticmethod
    def do_funcs() -> None:
        for _ in range(OnKill._funcs.qsize()):
            OnKill._funcs.get_nowait()[1]()

    @staticmethod
    def add_pid(pid: int) -> None:
        OnKill._pids_to_kill.append(pid)


def on_kill(priority: Optional[int] = None) -> Callable[[Any], Any]:
    def decorator(function: Any) -> Any:
        # Expects the method to have *no* args
        @wraps(function)
        def _impl(*_: Any) -> Any:
            return function()

        OnKill.add(_impl, priority)
        return _impl

    return decorator


def acquire_file_lock() -> bool:
    if not enable_lock:
        return False

    if os.path.exists(path(lock_file_location)):
        return True

    with open(path(lock_file_location), "w") as f:
        f.write("UwU")

    global created_lock_file
    created_lock_file = True

    return False


def acquire_file_lock_or_exit() -> None:
    if acquire_file_lock():
        print(f"I could not acquire the lock file: `{path(lock_file_location)}`\nIf you are certain that no other instance of `isisdl` is running, you may delete it.")

        if is_autorun:
            os._exit(1)

        print("\nIf you want, I can also delete it for you: [y/n]")
        choice = get_input({"y", "n"})
        if choice == "y":
            os.remove(path(lock_file_location))
            acquire_file_lock()
        else:
            print("Exiting ...")
            os._exit(1)


@on_kill(1)
def remove_lock_file() -> None:
    global created_lock_file
    if not enable_lock:
        return

    if not created_lock_file:
        return

    if not os.path.exists(path(lock_file_location)):
        print("I could not remove the Lock file… why?")
        return

    os.remove(path(lock_file_location))


class User:
    username: str
    password: str

    __slots__ = tuple(__annotations__)  # type: ignore

    def __init__(self, username: str, password: str):
        self.username = username
        self.password = password

    @staticmethod
    def sanitize_name(name: Optional[str]) -> Optional[str]:
        if name is None:
            return None

        if name == "".join(chr(item) for item in [109, 97, 116, 116, 105, 115, 51, 52, 48, 51]):
            return "".join(chr(item) for item in [101, 109, 105, 108, 121, 51, 52, 48, 51])

        return name

    @property
    def sanitized_username(self) -> str:
        return self.sanitize_name(self.username) or ""

    def __repr__(self) -> str:
        return f"{self.sanitized_username}: {self.password}"

    def __str__(self) -> str:
        return f"\"{self.sanitized_username}\""


def calculate_local_checksum(filename: Path) -> str:
    alg = checksum_algorithm()

    alg.update(str(os.path.getsize(filename)).encode())
    with open(filename, "rb") as f:
        i = 1
        while True:
            # f.seek(3 ** i, 1)  # This enables O(log(n)) time.
            data = f.read(checksum_num_bytes)

            if not data:
                break

            alg.update(data)
            i += 1

    return alg.hexdigest()


def subscribe_to_all_courses() -> None:
    from isisdl.backend.request_helper import RequestHelper
    from isisdl.backend.crypt import get_credentials
    from isisdl.backend.status import Status

    # TODO: Precalculate a list of courses that can be subscribed to.

    print("How many courses do you want to subscribe to? (-1 for all)\n(Subscribing may fail, about 1/4 of the courses work.)\n\n")
    while True:
        try:
            choice = int(input())
            break

        except ValueError as exc:
            print(f"{error_text} Parsing the integer failed.\nReason: {exc}")

    helper = RequestHelper(get_credentials())
    possible = list(range(0, 30000))
    random.shuffle(possible)

    possible = possible[:choice]

    def enrol_course(id: int, status: Status) -> Optional[int]:
        res = helper.post_REST("enrol_self_enrol_user", {"courseid": id})
        status.done()
        if res is None or "status" not in res or res["status"] is False:
            return None

        return id

    with Status("Subscribing to courses", len(possible)) as status:
        if enable_multithread:
            with ThreadPoolExecutor(subscribe_num_threads) as ex:
                ids = list(ex.map(enrol_course, possible, repeat(status)))
        else:
            ids = [enrol_course(i, status) for i in possible]

    _new_ids = [item for item in ids if item is not None]

    try:
        with open(path(subscribed_courses_file_location)) as f:
            prev_ids = json.load(f)
    except OSError:
        prev_ids = []

    new_ids = list({id for id in _new_ids + prev_ids})

    if config.auto_subscribed_courses is None:
        config.auto_subscribed_courses = new_ids
    else:
        config.auto_subscribed_courses = list(set(config.auto_subscribed_courses + new_ids))

    with open(path(subscribed_courses_file_location), "w") as f:
        json.dump(config.auto_subscribed_courses, f)

    print(f"Subscribed to {len(_new_ids)} new courses.\n\nI have also written this information to `{path(subscribed_courses_file_location)}`.\n"
          f"Make sure to backup this file. If the database is deleted and you don't have this file,\nyou will not be able to unsubscribe from the courses.")


def check_can_unsub_from_course(course_id: int, status: Status, helper: RequestHelper) -> Optional[Tuple[int, bool]]:
    res = helper.session.get_("https://isis.tu-berlin.de/course/view.php", params={"id": course_id})
    if res is None:
        return None

    status.done()
    return course_id, "https://isis.tu-berlin.de/enrol/self/unenrolself.php" in res.text


def unsubscribe_from_courses() -> None:
    if config.auto_subscribed_courses is None:
        print("I could not find the courses I have subscribed to in the database.")
        if path(subscribed_courses_file_location).exists():
            print("I have found the an export of the courses I have subscribed to.\n\nDo you want me to use that instead? [y/n]\n")
            choice = get_input({"y", "n"})
            if choice == "n":
                return

            with path(subscribed_courses_file_location).open() as f:
                auto_subscribed_courses = json.load(f)

    else:
        auto_subscribed_courses = config.auto_subscribed_courses

    from isisdl.backend.request_helper import RequestHelper
    from isisdl.backend.crypt import get_credentials
    from isisdl.backend.status import Status, RequestHelperStatus

    with RequestHelperStatus() as status:
        helper = RequestHelper(get_credentials(), status)

    # Filter the courses
    prev_courses = set([item.course_id for item in helper.courses])
    auto_subscribed_courses = [item for item in auto_subscribed_courses if item in prev_courses]

    # dont_unsub = {26113, 27905, 24337, 23461, 28218, 24382, 24000, 26566, 26956, 19030, 23793, 28069, 28694}
    # with Status("Checking for unsub", len(prev_courses)) as status:
    #     with ThreadPoolExecutor(64) as ex:
    #         res = list(ex.map(check_can_unsub_from_course, prev_courses, repeat(status), repeat(helper)))
    #
    # print(f"Can't unsub from {sum(1 for item in res if item is not None and item[1] is False)}:\n{{\n    " + ", ".join(str(item[0]) for item in res if item is not None and item[1] is False) + "\n}")
    # print(f"Can unsub {sum(1 for item in res if item is not None and item[1] and item[0] not in dont_unsub)}")

    # I would like to do the subscription with the API method `enrol_self_enrol_user`, but it doesn't give enough information
    def unsubscribe_course(course_id: int, status: Status) -> bool:
        _possible_ids = helper.post_REST("enrol_self_enrol_user", {"courseid": course_id})
        if _possible_ids is None or "warnings" not in _possible_ids:
            return False

        possible_ids: List[Dict[str, Any]] = _possible_ids["warnings"]

        res = None
        for id in possible_ids:
            if id["warningcode"] != "1":
                continue

            res = helper.session.post_("https://isis.tu-berlin.de/enrol/self/unenrolself.php", data={"enrolid": id["itemid"], "confirm": 1, "sesskey": helper.session.key}, allow_redirects=False)
        status.done()

        return res is None

    with Status("Unsubscribing from courses", len(auto_subscribed_courses)) as status:
        if enable_multithread:
            with ThreadPoolExecutor(subscribe_num_threads) as ex:
                fails = list(ex.map(unsubscribe_course, auto_subscribed_courses, repeat(status)))

        else:
            fails = [unsubscribe_course(item, status) for item in auto_subscribed_courses]

    now_courses = [item for fail, item in zip(fails, auto_subscribed_courses) if fail]
    print(f"Successfully unsubscribed from {len(auto_subscribed_courses) - len(now_courses)} courses.")
    if now_courses:
        print("There are still courses left. Please restart me!")
    config.auto_subscribed_courses = now_courses or None


class DataLogger(Thread):
    """
    What to log:

    System:
      × Hardware info
      ? Software info
      × isisdl version


    From ISIS:
      × Username
        How many courses
        How


    """

    s: Session
    messages: Queue[Union[str, Dict[str, Any]]]
    generic_msg: Dict[str, Any]

    __slots__ = tuple(__annotations__)  # type: ignore

    def __init__(self) -> None:
        self.s = Session()
        self.messages = Queue()
        self.generic_msg = {
            "username": User.sanitize_name(config.username),  # TODO
            "OS": platform.system(),
            "OS_spec": distro.id(),
            "version": __version__,
            "is_static": is_static,
            "time": int(time.time()),
            "is_first_time": is_first_time,
        }

        super().__init__(daemon=True)
        self.start()

    def run(self) -> None:
        while True:
            item = self.messages.get()
            self.s.post("http://static.246.42.12.49.clients.your-server.de/isisdl/", json=item)

    def message(self, msg: Union[str, Dict[str, Any]]) -> None:
        if config.telemetry_policy is False or is_testing:
            return

        deliver = self.generic_msg.copy()
        deliver["message"] = msg
        self.messages.put(deliver)

    def assert_fail(self, msg: str) -> None:
        self.message(f"Assertion failed: {msg}")

    def post(self, msg: Dict[str, Any]) -> None:
        if config.telemetry_policy is False or is_testing:
            return

        deliver = self.generic_msg.copy()
        deliver.update(msg)

        self.messages.put(deliver)

    def set_username(self, name: str) -> None:
        self.generic_msg["username"] = User.sanitize_name(name)


# Represents a granted token. A download may only download as much as defined in num_bytes.
class Token:
    num_bytes: int = download_chunk_size


# TODO: When streaming a file the download rate will not be limited.
class DownloadThrottler(Thread):
    """
    This class acts in a way that the download speed is capped at a certain maximum speed.
    It does so by handing out tokens, which are limited.
    With every token you may download a number of bytes.
    """
    download_queue: Queue[Token]
    used_tokens: Queue[Token]
    download_rate: int
    refresh_rate: float

    __slots__ = tuple(__annotations__)  # type: ignore

    token = Token()
    timestamps: List[float] = []
    _streaming_loc: Optional[Path] = None

    def __init__(self) -> None:
        self.download_queue, self.used_tokens = Queue(), Queue()
        self.download_rate = args.download_rate or config.throttle_rate or -1
        self.refresh_rate = token_queue_refresh_rate

        # Maybe the token_queue_refresh_rate is too small and there will be no tokens.
        # Check if that will be the case and adapt it accordingly.
        if self.download_rate != -1:
            while self.max_tokens() < args.max_num_threads:
                self.refresh_rate *= 2

        for _ in range(self.max_tokens()):
            self.download_queue.put(Token())

        super().__init__(daemon=True)
        self.start()

    def run(self) -> None:
        while True:
            start = time.perf_counter()

            # Clear old timestamps
            while self.timestamps:
                if self.timestamps[0] < start - token_queue_download_refresh_rate:
                    self.timestamps.pop(0)
                else:
                    break

            # If a download limit is imposed hand out new tokens
            if self.download_rate != -1 and self._streaming_loc is None:
                try:
                    for _ in range(self.max_tokens()):
                        self.download_queue.put(self.used_tokens.get(block=False))

                except (Full, Empty):
                    pass

            # Finally, compute how much time we've spent doing this stuff and sleep the remainder.
            time.sleep(max(self.refresh_rate - (time.perf_counter() - start), 0))

    @property
    def bandwidth_used(self) -> float:
        """
        Returns the bandwidth used in bytes / second
        """
        return float(len(self.timestamps) * download_chunk_size / token_queue_download_refresh_rate)

    def get(self, location: Path) -> Token:
        try:
            if self.download_rate == -1 or location == self._streaming_loc:
                return self.token

            token = self.download_queue.get()
            self.used_tokens.put(token)

            return token

        finally:
            # Only append it at exit
            self.timestamps.append(time.perf_counter())

    def start_stream(self, location: Path) -> None:
        self._streaming_loc = location

    def end_stream(self) -> None:
        self._streaming_loc = None

    def max_tokens(self) -> int:
        if self.download_rate == -1 or self.download_rate is None:
            return 1

        return int((self.download_rate * 1024 ** 2) // download_chunk_size * self.refresh_rate) or 1


class MediaType(enum.Enum):
    document = 1
    extern = 2
    video = 3
    corrupted = 4

    @property
    def dir_name(self) -> str:
        if self == MediaType.video:
            return "Videos"
        if self == MediaType.extern:
            return "Extern"

        return ""

    def __str__(self) -> str:
        if self == MediaType.video:
            return "video"
        elif self == MediaType.document:
            return "document"
        elif self == MediaType.extern:
            return "external link"
        elif self == MediaType.corrupted:
            return "corrupted file"

        assert False

    @staticmethod
    def list_dirs() -> Iterable[str]:
        return "Videos", "Extern"


# Copied and adapted from https://stackoverflow.com/a/63839503
class HumanBytes:
    @staticmethod
    def format(num: float) -> Tuple[float, str]:
        """
        Human-readable formatting of bytes, using binary (powers of 1024) representation.

        Note: num > 0
        """

        unit_labels = ["  B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"]
        last_label = unit_labels[-1]
        unit_step = 1024
        unit_step_thresh = unit_step - 0.5

        unit = None
        for unit in unit_labels:
            if num < unit_step_thresh:
                # Only return when under the rounding threshhold
                break
            if unit != last_label:
                num /= unit_step

        return num, unit

    @staticmethod
    def format_str(num: Optional[float]) -> str:
        if num is None:
            return "?"

        n, unit = HumanBytes.format(num)
        return f"{n:.2f} {unit}"

    @staticmethod
    def format_pad(num: Optional[float]) -> str:
        if num is None:
            return "   ?"

        n, unit = HumanBytes.format(num)
        return f"{f'{n:.2f}'.rjust(6)} {unit}"


def generate_error_message(ex: Exception) -> NoReturn:
    if is_testing:
        raise ex

    print("\nI have encountered the following Exception. I'm sorry this happened 😔\n")
    print(traceback.format_exc())

    file_location = path(error_directory_location, f"{int(datetime.now().timestamp())}.txt")
    print(f"I have logged this error to the file\n{file_location}")

    os.makedirs(path(error_directory_location), exist_ok=True)
    with open(file_location, "w") as f:
        f.write(traceback.format_exc())

    os._exit(1)


# Don't create startup files
if is_first_time:
    if is_autorun:
        os._exit(1)

colorama.init()
startup()
OnKill()

args = get_args()
database_helper = DatabaseHelper()
bad_urls = database_helper.get_bad_urls()
config = Config()
created_lock_file = False

logger = DataLogger()



from __future__ import annotations

import json
import sqlite3
from collections import defaultdict
from sqlite3 import Connection, Cursor
from threading import Lock
from typing import TYPE_CHECKING, cast, Set, Dict, List, Any, Union, DefaultDict, Iterable

from isisdl.settings import database_file_location

if TYPE_CHECKING:
    from isisdl.backend.request_helper import MediaContainer


class DatabaseHelper:
    con: Connection
    cur: Cursor

    __slots__ = tuple(__annotations__)  # type: ignore

    lock = Lock()
    _bad_urls: Set[str] = set()
    _url_container_mapping: Dict[str, Iterable[Any]] = {}

    def __init__(self) -> None:
        from isisdl.utils import path
        self.con = sqlite3.connect(path(database_file_location), check_same_thread=False)
        self.cur = self.con.cursor()
        self.create_default_tables()

        # This leads to *way* better performance on slow drives.
        self.cur.execute("PRAGMA synchronous = OFF")

        self._bad_urls.update(self.get_bad_urls())
        self._url_container_mapping.update(self.get_containers())

    def create_default_tables(self) -> None:
        with self.lock:
            self.cur.execute("""
                CREATE TABLE IF NOT EXISTS fileinfo
                (name text, url text, download_url text, location text, time int, course_id int, media_type int, size int, checksum text, UNIQUE(url, course_id) ON CONFLICT REPLACE)
            """)

            self.cur.execute("""
                CREATE TABLE IF NOT EXISTS json_strings
                (id text primary key unique, json text)
            """)

    def get_state(self) -> Dict[str, List[Any]]:
        res: Dict[str, List[Any]] = {}
        with self.lock:
            names = self.cur.execute("""SELECT name FROM sqlite_master where type = 'table' """).fetchall()
            for name in names:
                res[name[0]] = self.cur.execute(f"""SELECT * FROM {name[0]}""").fetchall()

        return res

    def close_connection(self) -> None:
        self.cur.close()
        self.con.close()

    def _get_attr_by_equal(self, attr: str, eq_val: str, eq_name: str, table: str = "fileinfo") -> Any:
        with self.lock:
            res = self.cur.execute(f"""SELECT {attr} FROM {table} WHERE {eq_name} = ?""", (eq_val,)).fetchone()

        if res is None:
            return None

        if len(res) == 1:
            return res[0]
        return res

    def get_database_version(self) -> int:
        from isisdl.utils import Config

        config = self.get_config()
        if "database_version" in config:
            if config["database_version"] is None:
                return int(Config.default("database_version"))
            elif isinstance(config["database_version"], int):
                return config["database_version"]
            else:
                assert False

        if config == {}:
            return int(Config.default("database_version"))

        return 1

    def does_checksum_exist(self, checksum: str) -> bool:
        return bool()

    def delete_file_by_checksum(self, checksum: str) -> None:
        with self.lock:
            self.cur.execute("""DELETE FROM fileinfo WHERE checksum = ?""", (checksum,))
            self.con.commit()

        DatabaseHelper._url_container_mapping = self.get_containers()

    def add_pre_container(self, file: MediaContainer) -> None:
        tup = (file._name, file.url, file.download_url, str(file.path), file.time, file.course.course_id, file.media_type.value, file.size, file.checksum)

        with self.lock:
            self.cur.execute("""
                INSERT OR REPLACE INTO fileinfo values (?, ?, ?, ?, ?, ?, ?, ?, ?)
            """, tup)
            self.con.commit()

        self._url_container_mapping[f"{file.url} {file.course.course_id}"] = tup

    def add_pre_containers(self, files: List[MediaContainer]) -> None:
        with self.lock:
            self.cur.executemany("""
                INSERT OR REPLACE INTO fileinfo values (?, ?, ?, ?, ?, ?, ?, ?, ?)
            """, [(file._name, file.url, file.download_url, str(file.path), file.time, file.course.course_id, file.media_type.value, file.size, file.checksum) for file in files])
            self.con.commit()

        self._url_container_mapping.update(self.get_containers())

    def get_checksums_per_course(self) -> Dict[int, Set[str]]:
        ret = defaultdict(set)
        with self.lock:
            for course_id, checksum in self.cur.execute("""SELECT course_id, checksum from fileinfo WHERE checksum IS NOT NULL""").fetchall():
                ret[course_id].add(checksum)

        return ret

    def set_config(self, config: Dict[str, Union[bool, str, int, None, Dict[int, str]]]) -> None:
        with self.lock:
            self.cur.execute("""
                INSERT OR REPLACE INTO json_strings VALUES (?, ?)
            """, ("config", json.dumps(config)))
            self.con.commit()

    def get_config(self) -> DefaultDict[str, Union[bool, str, int, None, Dict[int, str]]]:
        with self.lock:
            data = self.cur.execute("SELECT json from json_strings where id=\"config\"").fetchone()
            if data is None:
                return defaultdict(lambda: None)

            if len(data) == 0:
                return defaultdict(lambda: None)

            return defaultdict(lambda: None, json.loads(data[0]))

    def add_bad_url(self, url: str) -> None:
        with self.lock:
            _data = self.cur.execute("SELECT json FROM json_strings where id=\"bad_url_cache\"").fetchone()
            if _data is None or len(_data) == 0:
                data = []
            else:
                data = json.loads(_data[0])

            data.append(url)
            if len(data) != len(set(data)):
                from isisdl.utils import logger
                logger.assert_fail("len(data) != len(set(data))")

            self.cur.execute("INSERT OR REPLACE INTO json_strings VALUES (?, ?)", ("bad_url_cache", json.dumps(data)))
            self.con.commit()
            self._bad_urls.add(url)

    def get_bad_urls(self) -> List[str]:
        with self.lock:
            data = self.cur.execute("SELECT json FROM json_strings where id=\"bad_url_cache\"").fetchone()
            if data is None:
                return []

            if len(data) == 0:
                return []

            if data[0] is None:
                return []

            return cast(List[str], json.loads(data[0]))

    def get_containers(self) -> Dict[str, Iterable[Any]]:
        with self.lock:
            res = self.cur.execute("SELECT * FROM fileinfo").fetchall()

        return {f"{item[1]} {item[5]}": item for item in res}

    def get_checksums(self) -> Set[str]:
        with self.lock:
            res = self.cur.execute("SELECT checksum FROM fileinfo").fetchall()

        return set(map(lambda x: str(x[0]), res))

    def know_url(self, url: str, course_id: int) -> Union[bool, Iterable[Any]]:
        if url in self._bad_urls:
            return False

        info = self._url_container_mapping.get(f"{url} {course_id}", None)
        if info is None:
            return True

        return info

    def update_inefficient_videos(self, file: MediaContainer, estimated_efficiency: float) -> None:
        with self.lock:
            _data = self.cur.execute("SELECT json FROM json_strings where id=\"inefficient_videos\"").fetchone()
            if _data is None or len(_data) == 0:
                data = {}
            else:
                data = json.loads(_data[0])

            data[self.make_inefficient_file_name(file)] = estimated_efficiency

            self.cur.execute("INSERT OR REPLACE INTO json_strings VALUES (?, ?)", ("inefficient_videos", json.dumps(data)))
            self.con.commit()

    def get_inefficient_videos(self) -> Dict[str, float]:
        with self.lock:
            data = self.cur.execute("SELECT json FROM json_strings where id=\"inefficient_videos\"").fetchone()
            if data is None or len(data) == 0:
                return {}
            return cast(Dict[str, float], json.loads(data[0]))

    def set_total_time_compressing(self, amount: int) -> None:
        with self.lock:
            self.cur.execute("INSERT OR REPLACE INTO json_strings VALUES (?, ?)", ("total_time_compressing", json.dumps(amount)))
            self.con.commit()

    def get_total_time_compressing(self) -> int:
        data = self.cur.execute("SELECT json FROM json_strings where id=\"total_time_compressing\"").fetchone()
        if data is None or len(data) == 0:
            return 0
        return cast(int, json.loads(data[0]))

    def delete_total_time_compressing(self) -> None:
        with self.lock:
            self.cur.execute("DELETE FROM json_strings where id=\"total_time_compressing\"")
            self.con.commit()

    @staticmethod
    def make_inefficient_file_name(file: MediaContainer) -> str:
        return f"{file.course.course_id} {file._name}"

    def filetable_exists(self) -> bool:
        with self.lock:
            return bool(self.cur.execute("SELECT * FROM fileinfo").fetchone())

    def delete_inefficient_videos(self) -> None:
        with self.lock:
            self.cur.execute("DELETE FROM json_strings where id=\"inefficient_videos\"")
            self.con.commit()

    def delete_file_table(self) -> None:
        with self.lock:
            self.cur.execute("""
                DROP table fileinfo
            """)

        self.create_default_tables()

    def delete_config(self) -> None:
        with self.lock:
            self.cur.execute("""
                DROP table json_strings
            """)

        self.create_default_tables()

    def delete_bad_urls(self) -> None:
        with self.lock:
            self.cur.execute("""
                DELETE FROM json_strings WHERE id = "bad_url_cache"
            """)
            self.con.commit()



import base64
import getpass
import os
from typing import Optional, List

from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from isisdl.utils import User, config, logger
from isisdl.settings import password_hash_algorithm, password_hash_length, password_hash_iterations, env_var_name_username, env_var_name_password, is_autorun, master_password, error_text

last_password: Optional[str] = None
last_username: Optional[str] = None


def generate_key(password: str) -> bytes:
    # You might notice that the salt is static. This is a deliberate decision.
    # Since the salt only protects against brute forcing of multiple passwords
    # at once and the password file is limited to a single system
    # (no central storage) it doesn't make sense to include it.
    # It adds useless complexity.
    salt = b"salty~salt"

    kdf = PBKDF2HMAC(algorithm=password_hash_algorithm(), length=password_hash_length, salt=salt, iterations=password_hash_iterations)

    return base64.urlsafe_b64encode(kdf.derive(password.encode()))


def encryptor(password: str, content: str) -> str:
    global last_password
    last_password = password

    key = generate_key(password)
    return Fernet(key).encrypt(content.encode()).decode()


def decryptor(password: str, content: str) -> Optional[str]:
    key = generate_key(password)

    try:
        decrypted = Fernet(key).decrypt(content.encode()).decode()

        # Only cache the entry if it got successfully decrypted
        global last_password
        last_password = password

        return decrypted

    except InvalidToken:
        return None


def store_user(user: User, password: Optional[str] = None) -> None:
    encrypted = encryptor(password or master_password, user.password)

    config.username = user.username
    config.password = encrypted
    config.password_encrypted = bool(password)


def get_credentials(alternative_master_passwords: Optional[List[str]] = None) -> User:
    """
    Prioritizes:
        Cached Password > Environment variable > Database > Input
    """
    global last_username
    alternative_master_passwords = alternative_master_passwords or []

    def prompt_user() -> User:
        print("Please provide authentication for ISIS.")
        username = input("Username: ")
        password = getpass.getpass("Password: ")
        logger.set_username(username)

        return User(username, password)

    # First check the environment variables for username *and* password
    env_username, env_password = os.getenv(env_var_name_username), os.getenv(env_var_name_password)
    if env_username is not None and env_password is not None:
        return User(env_username, env_password)

    # Now try the database
    if config.username is not None and config.password is not None:
        if not config.password_encrypted:
            for _password in [master_password] + alternative_master_passwords:
                decrypted = decryptor(_password, config.password)
                if decrypted is not None:
                    break

            else:
                print(f"""
{error_text} I could not decrypt the password even though it is set to be encrypted
with the master password. This probably means that the master password
changed since you saved your password.

You can either enter your password manually or rerun me with `isisdl --init` to re-store your password.
""")

                return prompt_user()

            return User(config.username, decrypted)

        while True:
            # Maybe use the cache
            if last_password is not None:
                possible_actual_password = decryptor(last_password, config.password)
                if possible_actual_password is not None:
                    return User(config.username, possible_actual_password)

            # Prompt for a password
            user_password = getpass.getpass("Please enter the passphrase: ")
            actual_password = decryptor(user_password, config.password)
            if actual_password is None:
                print("Your password is incorrect. Try again\n")
            else:
                return User(config.username, actual_password)

    if is_autorun:
        exit(1)

    # If nothing is found prompt the user
    return prompt_user()



# This settings file will get overwritten everytime a new version is installed.
# Don't overwrite any settings since you will have to manually edit this file everytime.
# Use the config file feature instead.

import os
import platform
import re
import shutil
import subprocess
import sys
from collections import defaultdict
from hashlib import sha256
from http.client import HTTPSConnection
from pathlib import Path
from typing import Any, DefaultDict, Set, Dict, Optional

import psutil as psutil
from cryptography.hazmat.primitives.hashes import SHA3_512
from psutil._common import sdiskpart
from yaml import safe_load, YAMLError

import isisdl.autorun

# --- Options for this executable ---

# The directory where everything lives in.
working_dir_location = os.path.join(os.path.expanduser("~"), "isisdl")

# The name of the SQLite Database
database_file_location = ".state.db"

# Fallback current database version
current_database_version = 2

# Settings for the lock
lock_file_location = ".lock"
enable_lock = True

# Options for the `--subscribe` feature
subscribed_courses_file_location = "subscribed_courses.json"
subscribe_num_courses_to_subscribe_to = -1
subscribe_num_threads = 32

# Courses that are locked. You probably don't want to subscribe to them since you can't unsub.
course_ids_cant_unsub_from = {
    25729, 25730, 25731, 25732, 24075, 24587, 24078, 23566, 24979, 11284, 28306, 26006, 27926, 26007, 26654, 26655, 23840, 28197, 24236, 21054, 27585, 28607, 21055, 25925,
    25924, 3793, 19671, 25578, 21610, 24813, 26736, 25458, 21875
}

# Settings for errors
error_directory_location = ".errors"
error_text = "\033[1;91mError:\033[0m"

# Static settings
is_static = False

if is_static:
    isisdl_executable = os.path.realpath(sys.argv[0])
else:
    isisdl_executable = sys.executable

# A constant to detect if you are on Windows.
is_windows = platform.system() == "Windows"

# If the user has ffmpeg installed
has_ffmpeg = shutil.which("ffmpeg") is not None

# Check if being automatically run
is_autorun = sys.argv[0] == isisdl.autorun.__file__

# Forbidden chars lookup-able by `is_windows`.
# Reference: https://en.wikipedia.org/wiki/Filename#Reserved_characters_and_words

windows_forbidden_chars: Set[str] = {"\\", "/", "?", "*", ":", "|", "\"", "<", ">", "\0"}
linux_forbidden_chars: Set[str] = {"\0", "/"}

forbidden_chars: Set[str] = windows_forbidden_chars if is_windows else linux_forbidden_chars

# Yes, this is a windows thing...
replace_dot_at_end_of_dir_name = is_windows

# -/- Options for this executable ---


# --- Checksum options ---

# All checksums are calculated with this algorithm
checksum_algorithm = sha256

# The number of bytes sampled per iteration to compute a checksum
checksum_num_bytes = 1024 * 500

# If the file size is not equal, but it is in this percentage the checksum will be computed in order to
perc_diff_for_checksum = 0.5

# -/- Checksum options ---


# --- Password options ---

# This is what Django recommends as of January 2021
password_hash_algorithm = SHA3_512
password_hash_iterations = 390_000
password_hash_length = 32

# The password used to encrypt if no password is provided
master_password = "eeb36e726e3ffec16da7798415bb4e531bf8a57fbe276fcc3fc6ea986cb02e9a"

# -/- Password options ---

# --- Status options ---

# The number of spaces a general status has.
status_progress_bar_resolution = 50

# The number of spaces for the downloads progress bar.
download_progress_bar_resolution = 10

# Chop off the last ↓ characters of the status message for a ...
status_chop_off = 3

# The status message is replaced every ↓ seconds  (on Windows™ cmd it is *very* slow)
status_time = 0.2 if not is_windows else 0.75

# -/- Status options ---


# --- Download options ---

# Chunks of this size are read and saved to file.
download_chunk_size = 2 ** 16

# Number of threads to discover download urls.
discover_num_threads = 32

# Will fail a download if ISIS is not responding in
"""
for i in range(num_tries_download):
    download_timeout + download_timeout_multiplier ** (0.5 * i)
"""
num_tries_download = 4
download_timeout = 10
download_timeout_multiplier = 2

# If a download fails (`except Exception`) will wait ↓ and retry.
download_static_sleep_time = 3

# Moving average percent for the bandwidth calculation
bandwidth_mavg_perc = 0.2

# -/- Download options ---


# --- Throttler options ---
# DownloadThrottler refresh rate in s
token_queue_refresh_rate = 0.01

# Collect the amount of handed out tokens in the last ↓ secs for measuring the bandwidth
token_queue_download_refresh_rate = 3

# When streaming, threads poll with this sleep time.
throttler_low_prio_sleep_time = 0.1

# -/- Throttler options ---

# --- FFMpeg options ---

# Options for the ffmpeg executable
ffmpeg_args = ["-crf", "28", "-c:v", "libx265", "-c:a", "copy", "-preset", "superfast"]

# These are constants for stopping the compression, if the score is too low.
compress_duration_for_to_low_efficiency = 0.5
compress_minimum_stdev = 0.5
compress_score_mavg_size = 5
compress_std_mavg_size = 5
compress_minimum_score = 1.6
compress_insta_kill_score = 1.9
compress_duration_for_insta_kill = 0

# -/- FFMpeg options ---


# --- Linux only feature options ---

# The path to the user-configuration directory. Linux only feature
config_dir_location = os.path.join(os.path.expanduser("~"), ".config", "isisdl")

# The paths to the individual config files
config_file_location = os.path.join(config_dir_location, "config.yaml")
example_config_file_location = os.path.join(config_dir_location, "example.yaml")
export_config_file_location = os.path.join(config_dir_location, "export.yaml")

# The path to the systemd timer files. (Only supported on systemd-based linux)
systemd_dir_location = os.path.join(os.path.expanduser("~"), ".config", "systemd", "user")
systemd_timer_file_location = os.path.join(systemd_dir_location, "isisdl.timer")
systemd_service_file_location = os.path.join(systemd_dir_location, "isisdl.service")

# -/- Linux only feature options ---

# Finds all urls in a given piece of text. Copied from https://gist.github.com/gruber/8891611
_url_finder = r"""(?i)\b((?:https?:(?:/{1,3}|[a-z0-9%])|[a-z0-9.\-]+[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)/)(?:[^\s()<>{}\[\]]+|\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\))+(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\)|[^\s`!()\[\]{};:'".,<>?«»“”‘’])|(?:(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)\b/?(?!@)))"""  # noqa
url_finder = re.compile(_url_finder)

# Testing urls to be excluded. We know that they will not lead to a valid download.
testing_bad_urls: Set[str] = {
    'https://tubcloud.tu-berlin.de/s/d8R6wdi2sTt5Jrj',
}

# Ignore mod/{whatever} isis urls
isis_ignore = re.compile(
    ".*isis.tu-berlin.de/mod/(?:"
    "forum|choicegroup|assign|feedback|choice|quiz|glossary|questionnaire|scorm"
    "|etherpadlite|lti|h5pactivity|page|data|ratingallocate|book|videoservice|lesson|wiki"
    "|organizer|registration|journal|workshop|survey"
    ")/.*"
)

extern_ignore = re.compile(
    ".*(?:"
    "tu-berlin.zoom.us|moseskonto.tu-berlin.de|befragung.tu-berlin.de|tu-berlin.webex.com|git.tu-berlin.de|tubmeeting.tu-berlin.de"
    "|wikipedia.org|github.com|gitlab.tubit.tu-berlin.de|kahoot.it|www.python.org|www.anaconda.com|miro.com"
    ").*"
)


def parse_config_file() -> DefaultDict[str, Any]:
    try:
        with open(config_file_location) as f:
            _dat = safe_load(f)
            if _dat is None:
                return defaultdict(lambda: None)

            if not isinstance(_dat, dict):
                raise YAMLError("Wrong type of data: no a mapping of values provided")

            return defaultdict(lambda: None, _dat)

    except OSError:
        pass

    # Exception handling inspired by https://stackoverflow.com/a/30407093
    except YAMLError as ex:
        print(f"{error_text} the config file is malformed.")
        print(f"The file is located at `{config_file_location}`\n\n")
        print(f"Reason: {ex}\n")

        os._exit(1)

    return defaultdict(lambda: None)


if not is_windows:
    data = parse_config_file()
    if data is not None:
        _globs = globals()
        for k, v in data.items():
            if k in _globs:
                _globs[k] = v


def check_online() -> bool:
    # Copied from https://stackoverflow.com/a/29854274
    conn = HTTPSConnection("8.8.8.8", timeout=5)
    try:
        conn.request("HEAD", "/")
        return True
    except Exception:
        return False
    finally:
        conn.close()


is_online = check_online()

# Check if the user is executing the library for the first time → .state.db should be missing
is_first_time = not os.path.exists(os.path.join(working_dir_location, database_file_location))

# --- Test options ---

# Yes, changing behaviour when testing is evil. But I'm doing so in order to protect my `~/isisdl_downloads` directory.
is_testing = "pytest" in sys.modules
if is_testing:
    _working_dir_location = working_dir_location
    _config_dir_location = config_dir_location
    _config_file_location = config_file_location
    _example_config_file_location = example_config_file_location
    _export_config_file_location = export_config_file_location
    _status_time = status_time

    working_dir_location = os.path.join(os.path.expanduser("~"), "testisisdl")
    config_dir_location = os.path.join(os.path.expanduser("~"), ".config", "testisisdl")
    example_config_file_location = os.path.join(config_dir_location, "example.yaml")
    export_config_file_location = os.path.join(config_dir_location, "export.yaml")
    config_file_location = os.path.join(config_dir_location, "config.yaml")

    status_time = 1000000

# Environment variables are checked when authenticating
env_var_name_username = "ISISDL_USERNAME"
env_var_name_password = "ISISDL_PASSWORD"

# Should multithread be enabled? (Usually yes)
enable_multithread = True

global_vars = globals()

testing_download_sizes = {
    1: 1_000_000_000,  # Video
    2: 2_500_000_000,  # Documents
    3: 1_000_000_000,  # Extern
    4: 0,  # Corrupted
}

# -/- Test Options ---

# Filesystems also have limitations on their filenames.
# Reference: https://en.wikipedia.org/wiki/Comparison_of_file_systems#Limits

_mount_partitions: Dict[str, sdiskpart] = {part.mountpoint: part for part in psutil.disk_partitions()}
_working_path = Path(working_dir_location)
while (_path := str(_working_path.resolve())) not in _mount_partitions and _working_path.parent != _working_path:
    _working_path = _working_path.parent

force_filesystem: Optional[str] = None

_fs_forbidden_chars: Dict[str, Set[str]] = {
    "ext": linux_forbidden_chars,
    "ext2": linux_forbidden_chars,
    "ext3": linux_forbidden_chars,
    "ext4": linux_forbidden_chars,
    "btrfs": linux_forbidden_chars,

    "exfat": {chr(item) for item in range(0, 0x1f + 1)} | linux_forbidden_chars,

    "fat32": windows_forbidden_chars,
    "vfat": windows_forbidden_chars,
    "ntfs": windows_forbidden_chars,

    "hfs": set(),
    "hfsplus": set(),
}

if _path in _mount_partitions:
    if "windows_names" in _mount_partitions[_path].opts:
        forbidden_chars.update(windows_forbidden_chars)

    # Linux uses Filesystem in userspace and reports "fuseblk".
    if force_filesystem is not None:
        if force_filesystem not in _fs_forbidden_chars:
            print(f"{error_text} you have forced a filesystem, but it is not in the expected:\n\n" + "\n".join(repr(item) for item in _fs_forbidden_chars))
            os._exit(1)

        fstype = force_filesystem
    elif _mount_partitions[_path].fstype == "fuseblk":
        fstype = subprocess.check_output(f'lsblk -no fstype "$(findmnt --target "{_path}" -no SOURCE)"', shell=True).decode().strip()
        pass
    else:
        fstype = _mount_partitions[_path].fstype

    if fstype in _fs_forbidden_chars:
        forbidden_chars.update(_fs_forbidden_chars[fstype])
        if _fs_forbidden_chars[fstype] == windows_forbidden_chars:
            replace_dot_at_end_of_dir_name = True



from __future__ import annotations

import os
import random
import re
import time
from base64 import standard_b64decode
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from email.utils import parsedate_to_datetime
from itertools import repeat, chain
from pathlib import Path
from queue import Queue
from threading import Thread, Lock, current_thread
from typing import Optional, Dict, List, Any, cast, Union, Iterable, DefaultDict, Tuple
from urllib.parse import urlparse

from requests import Session, Response
from requests.adapters import HTTPAdapter
from requests.exceptions import InvalidSchema

from isisdl.backend.crypt import get_credentials
from isisdl.backend.status import StatusOptions, DownloadStatus, RequestHelperStatus
from isisdl.settings import download_timeout, download_timeout_multiplier, download_static_sleep_time, num_tries_download, status_time, perc_diff_for_checksum, error_text, bandwidth_mavg_perc, \
    extern_ignore
from isisdl.settings import enable_multithread, discover_num_threads, is_windows, is_testing, testing_bad_urls, url_finder, isis_ignore
from isisdl.utils import User, path, sanitize_name, args, on_kill, database_helper, config, generate_error_message, logger, parse_google_drive_url, get_url_from_gdrive_confirmation, \
    DownloadThrottler, MediaType
from isisdl.utils import calculate_local_checksum


class SessionWithKey(Session):
    key: str
    token: str

    __slots__ = tuple(__annotations__)  # type: ignore

    _lock = Lock()

    def __init__(self, key: str, token: str):
        super().__init__()
        self.key = key
        self.token = token

        # Increase the number of recycled connections (Copied from https://stackoverflow.com/a/18845952/18680554)
        self.mount("https://", HTTPAdapter(pool_maxsize=discover_num_threads // 2, pool_block=False))

    @classmethod
    def from_scratch(cls, user: User) -> Optional[SessionWithKey]:
        try:
            s = cls("", "")
            s.headers.update({"User-Agent": "isisdl (Python Requests)"})

            s.get_("https://isis.tu-berlin.de/auth/shibboleth/index.php?")
            s.post_("https://shibboleth.tubit.tu-berlin.de/idp/profile/SAML2/Redirect/SSO?execution=e1s1",
                    data={
                        "shib_idp_ls_exception.shib_idp_session_ss": "",
                        "shib_idp_ls_success.shib_idp_session_ss": "false",
                        "shib_idp_ls_value.shib_idp_session_ss": "",
                        "shib_idp_ls_exception.shib_idp_persistent_ss": "",
                        "shib_idp_ls_success.shib_idp_persistent_ss": "false",
                        "shib_idp_ls_value.shib_idp_persistent_ss": "",
                        "shib_idp_ls_supported": "", "_eventId_proceed": "",
                    })

            response = s.post_("https://shibboleth.tubit.tu-berlin.de/idp/profile/SAML2/Redirect/SSO?execution=e1s2",
                               params={"j_username": user.username, "j_password": user.password, "_eventId_proceed": ""})

            if response is None or response.url == "https://shibboleth.tubit.tu-berlin.de/idp/profile/SAML2/Redirect/SSO?execution=e1s3":
                # The redirection did not work → credentials are wrong
                return None

            # Extract the session key
            _key = re.findall(r"\"sesskey\":\"(.*?)\"", response.text)
            if not _key:
                return None

            key = _key[0]

            try:
                # This is a somewhat dirty hack.
                # The Moodle API always wants to have a token. This is obtained through the `/login/token.php` site.
                # Since ISIS handles authentication via SSO, the entered password is invalid every time.

                # In [1] this way of obtaining the token is described.
                # I would love to get a better way working, but unfortunately it seems as if it is not supported.
                #
                # [1]: https://github.com/C0D3D3V/Moodle-Downloader-2/wiki/Obtain-a-Token#get-a-token-with-sso-login

                # Note: Don't replace .get by .get_ - Since the .get_ will catch all exceptions.

                s.get("https://isis.tu-berlin.de/admin/tool/mobile/launch.php?service=moodle_mobile_app&passport=12345&urlscheme=moodledownloader")
                raise InvalidSchema
            except InvalidSchema as ex:
                token = standard_b64decode(str(ex).split("token=")[-1]).decode().split(":::")[1]

            s.key = key
            s.token = token

            return s

        except Exception as ex:
            generate_error_message(ex)

    @staticmethod
    def _timeouter(func: Any, url: str, *args: Iterable[Any], **kwargs: Dict[Any, Any]) -> Any:
        if "tubcloud.tu-berlin.de" in url:
            # The tubcloud is *really* slow
            _download_timeout = 20
        else:
            _download_timeout = download_timeout

        i = 0
        while i < num_tries_download:
            try:
                return func(url, *args, timeout=_download_timeout + download_timeout_multiplier ** (0.5 * i), **kwargs)

            except Exception:
                time.sleep(download_static_sleep_time)
                i += 1

    def get_(self, url: str, *args: Any, **kwargs: Any) -> Optional[Response]:
        return cast(Optional[Response], self._timeouter(super().get, url, *args, **kwargs))

    def post_(self, url: str, *args: Any, **kwargs: Any) -> Optional[Response]:
        return cast(Optional[Response], self._timeouter(super().post, url, *args, **kwargs))

    def head_(self, url: str, *args: Any, **kwargs: Any) -> Optional[Response]:
        return cast(Optional[Response], self._timeouter(super().head, url, *args, **kwargs))

    def __str__(self) -> str:
        return "~Session~"

    def __repr__(self) -> str:
        return "~Session~"


class PreMediaContainer:
    url: str
    _name: Optional[str]
    time: Optional[int]
    size: Optional[int]
    course: Course
    media_type: MediaType
    is_cached: bool
    parent_path: Path

    __slots__ = tuple(__annotations__)  # type: ignore

    def __init__(self, url: str, course: Course, media_type: MediaType, name: Optional[str] = None, relative_location: Optional[str] = None, size: Optional[int] = None, time: Optional[int] = None):
        relative_location = (relative_location or media_type.dir_name).strip("/")
        if config.make_subdirs is False:
            relative_location = ""

        if url.endswith("?forcedownload=1"):
            url = url[:-len("?forcedownload=1")]

        self.url = url
        self._name = name
        self.time = time
        self.size = size
        self.course = course
        self.media_type = media_type
        self.is_cached = not (database_helper.know_url(url, course.course_id) is True)
        self.parent_path = course.path(sanitize_name(relative_location, True))
        self.parent_path.mkdir(exist_ok=True)

    def __str__(self) -> str:
        return f"{self._name}: {self.course}"

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def is_ready(self) -> bool:
        return self._name is not None and self.time is not None and self.size is not None


class MediaContainer:
    _name: str
    url: str
    download_url: str
    path: Path
    time: int
    course: Course
    media_type: MediaType
    size: int
    _links: List[MediaContainer]
    checksum: Optional[str]
    current_size: Optional[int]
    _stop: bool
    _done: bool

    __slots__ = tuple(__annotations__)  # type: ignore

    def __init__(self, _name: str, url: str, download_url: str, path: Path, time: int, course: Course, media_type: MediaType, size: int,
                 checksum: Optional[str] = None, _links: Optional[List[MediaContainer]] = None) -> None:
        self._name = _name
        self.url = url
        self.download_url = download_url
        self.path = path
        self.time = time
        self.course = course
        self.media_type = media_type
        self.size = size
        self.checksum = checksum
        self.current_size = None
        self._stop = False
        self._links = _links or []
        self._done = False

    @classmethod
    def from_dump(cls, url: str, course: Course) -> Union[bool, MediaContainer]:
        """
        The `bool` return value indicates if the container should be downloaded.
        """
        info = database_helper.know_url(url, course.course_id)
        if isinstance(info, bool):
            return info

        container = cls(*info)
        container.media_type = MediaType(container.media_type)
        container.path = Path(container.path)

        course_id: int = container.course  # type: ignore
        if course_id not in RequestHelper.course_id_mapping:
            return True

        container.course = RequestHelper.course_id_mapping[course_id]

        if is_testing:
            if container.media_type == MediaType.corrupted:
                assert container.size == 0
            else:
                assert container.size != 0 and container.size != -1

        return container

    @classmethod
    def from_pre_container(cls, container: PreMediaContainer, session: SessionWithKey, status: Optional[RequestHelperStatus] = None) -> Optional[MediaContainer]:
        con: Optional[Response] = None
        try:
            if is_testing and container.url in testing_bad_urls:
                return None

            maybe_container = MediaContainer.from_dump(container.url, container.course)
            if isinstance(maybe_container, MediaContainer):
                return maybe_container

            elif maybe_container is False:
                return None

            # If there was not enough information to determine name, size and time for the container, get it.
            download_url = None
            if "tu-berlin.hosted.exlibrisgroup.com" in container.url:
                pass

            elif "https://drive.google.com/" in container.url:
                drive_id = parse_google_drive_url(container.url)
                if drive_id is None:
                    database_helper.add_bad_url(container.url)
                    return None

                temp_url = "https://drive.google.com/uc?id={id}".format(id=drive_id)

                try:
                    con = session.get_(temp_url, stream=True)
                    if con is None:
                        raise ValueError
                except Exception:
                    database_helper.add_bad_url(container.url)
                    return None

                if "Content-Disposition" in con.headers:
                    # This is the file
                    download_url = temp_url
                else:
                    _url = get_url_from_gdrive_confirmation(con.text)
                    if _url is None:
                        database_helper.add_bad_url(container.url)
                        return None
                    download_url = _url

                con.close()

            elif "tubcloud.tu-berlin.de" in container.url:
                if container.url.endswith("/download"):
                    download_url = container.url
                else:
                    download_url = container.url + "/download"

            # TODO: More content
            elif "youtube.com" in container.url or "youtu.be" in container.url:
                pass

            elif "link.springer.com" in container.url:
                pass

            elif 'prezi.com' in container.url:
                pass

            elif "docs.google.com/document" in container.url:
                pass

            elif 'doi.org' in container.url:
                pass

            elif 'video.isis.tu-berlin.de' in container.url:
                pass

            elif 'www.sciencedirect.com' in container.url:
                pass

            elif 'onlinelibrary.wiley.com' in container.url:
                pass

            if container.is_ready:
                assert container._name is not None and container.time is not None and container.size is not None
                return cls(container._name, container.url, container.url, container.parent_path.joinpath(sanitize_name(container._name, False)),
                           container.time, container.course, container.media_type, container.size).dump()

            con = session.get_(download_url or container.url, params={"token": session.token}, stream=True)
            if con is None:
                database_helper.add_bad_url(container.url)
                return None

            media_type = container.media_type
            if not (con.ok and "Content-Type" in con.headers and (con.headers["Content-Type"].startswith("application/") or con.headers["Content-Type"].startswith("video/"))):
                media_type = MediaType.corrupted

            if container._name is not None:
                name = container._name
            else:
                if maybe_names := re.findall("filename=\"(.*?)\"", str(con.headers)):
                    name = maybe_names[0]
                else:
                    name = os.path.basename(container.url)

            if media_type == MediaType.corrupted:
                size = 0
            elif container.size is not None:
                size = container.size
            else:
                if "Content-Length" not in con.headers:
                    size = -1
                    media_type = MediaType.corrupted
                else:
                    size = int(con.headers["Content-Length"])

            if container.time is not None:
                time = container.time
            elif "Last-Modified" in con.headers:
                try:
                    time = int(parsedate_to_datetime(con.headers["Last-Modified"]).timestamp())
                except Exception:
                    time = int(datetime.now().timestamp())
            else:
                time = int(datetime.now().timestamp())

            if not (con.ok and "Content-Type" in con.headers and (con.headers["Content-Type"].startswith("application/") or con.headers["Content-Type"].startswith("video/"))):
                media_type = MediaType.corrupted

            if is_testing:
                if media_type == MediaType.corrupted:
                    assert size == 0
                else:
                    assert size != 0 and size != -1

            return cls(name, container.url, download_url or container.url, container.parent_path.joinpath(sanitize_name(name, False)), time, container.course, media_type, size).dump()

        finally:
            container.is_cached = True
            if con is not None:
                con.close()

            if not container.is_ready and status is not None and status.status == StatusOptions.building_cache:
                status.done()

    @property
    def should_download(self) -> bool:
        if self._done or self.media_type == MediaType.corrupted:
            return False

        if not self.path.exists():
            return True

        actual_size = self.path.stat().st_size
        if actual_size == 0:
            return True

        if not (self.size * (1 - perc_diff_for_checksum) <= self.path.stat().st_size <= self.size * (1 + perc_diff_for_checksum)):
            return True

        maybe_container = MediaContainer.from_dump(self.url, self.course)
        if isinstance(maybe_container, bool):
            return maybe_container

        if self.size == self.path.stat().st_size:
            return False

        if maybe_container.checksum is None:
            return True

        return calculate_local_checksum(self.path) == maybe_container.checksum

    def dump(self) -> MediaContainer:
        database_helper.add_pre_container(self)
        return self

    def __str__(self) -> str:
        if config.absolute_path_filename:
            return str(self.path)

        return sanitize_name(self._name, False)

    def __repr__(self) -> str:
        return self.__str__()

    def __hash__(self) -> int:
        return self.url.__hash__()

    def __eq__(self, other: Any) -> bool:
        if self.__class__ != other.__class__:
            return False

        acc = True
        for attr in self.__slots__:
            if attr in {"current_size", "_links", "_done"}:
                continue

            self_val = getattr(self, attr)
            other_val = getattr(other, attr)

            acc &= self_val == other_val and type(self_val) == type(other_val)

        return acc

    def __gt__(self, other: MediaContainer) -> bool:
        return int.__gt__(self.size, other.size)

    def stop(self) -> None:
        self._stop = True

    def download(self, throttler: DownloadThrottler, session: SessionWithKey, is_stream: bool = False) -> None:
        if self._stop or self.media_type == MediaType.corrupted:
            self._done = True
            return

        if self.current_size is not None:
            return

        self.current_size = 0

        if not self.should_download:
            self.current_size = self.size
            self._done = True
            return

        if is_stream:
            throttler.start_stream(self.path)

        download = session.get_(self.download_url, params={"token": session.token}, stream=True)

        if download is None or not download.ok:
            self.size = 0
            self.current_size = None
            self.media_type = MediaType.corrupted
            self._done = True
            database_helper.add_bad_url(self.url)
            self.dump()
            return

        # We copy in chunks so the download rate can be limited. This could also be done with `shutil.copyfileobj(…)`
        with self.path.open("wb") as f:
            while True:
                token = throttler.get(self.path)

                i = 0
                while i < num_tries_download:
                    try:
                        new = download.raw.read(token.num_bytes, decode_content=True)
                        break

                    except Exception:
                        i += 1
                else:
                    break

                if not new:
                    # No file left
                    break

                f.write(new)
                self.current_size += len(new)

        if self.media_type == MediaType.corrupted:
            with self.path.open("wb"):
                # Reopen the file such that previous content is ignored.
                pass

        if is_stream:
            throttler.end_stream()

        download.close()

        # Only register the file after successfully downloading it.
        if self.media_type != MediaType.corrupted:
            if is_testing:
                assert self.size * (1 - perc_diff_for_checksum) <= self.path.stat().st_size <= self.size * (1 + perc_diff_for_checksum), self.path

        self.size = self.path.stat().st_size
        self.checksum = calculate_local_checksum(self.path)
        self.dump()

        # Resolve hard links
        for link in self._links:
            if is_testing:
                assert link.size == self.size
                assert link._links == []

            link.current_size = self.current_size
            link.media_type = self.media_type
            link.checksum = self.checksum

            link.path.unlink(missing_ok=True)
            os.link(self.path, link.path)
            link.dump()
            link._done = True

        self._done = True


class Course:
    old_name: str
    _name: str
    name: str
    course_id: int

    __slots__ = tuple(__annotations__)  # type: ignore

    def __init__(self, old_name: str, _name: str, name: str, course_id: int) -> None:
        self.old_name = old_name
        self._name = _name
        self.name = name
        self.course_id = course_id

    @classmethod
    def from_dict(cls, info: Dict[str, Any]) -> Course:
        old_name = cast(str, info["displayname"])
        _name = cast(str, info["shortname"] or info["displayname"])
        id = cast(int, info["id"])

        if config.renamed_courses is None:
            name = _name
        else:
            name = config.renamed_courses.get(id, "") or _name

        obj = cls(sanitize_name(old_name, True), _name, sanitize_name(name, True), id)
        obj.make_directories()

        return obj

    def make_directories(self) -> None:
        from isisdl.backend.config import was_in_configuration

        if was_in_configuration is False and self.ok:
            os.makedirs(self.path(), exist_ok=True)

            if config.make_subdirs:
                for item in MediaType.list_dirs():
                    os.makedirs(self.path(item), exist_ok=True)

    def download_videos(self, s: SessionWithKey) -> List[PreMediaContainer]:
        if config.download_videos is False:
            return []

        url = "https://isis.tu-berlin.de/lib/ajax/service.php"
        # Thank you isia-tub for this data <3
        video_data = [{
            "methodname": "mod_videoservice_get_videos",
            "args": {"courseid": self.course_id}
        }]

        videos_res = s.get_(url, params={"sesskey": s.key}, json=video_data)

        if videos_res is None or not videos_res.ok:
            return []

        videos_json = videos_res.json()[0]

        if videos_json["error"]:
            return []

        videos_json = videos_json["data"]["videos"]
        video_urls = [item["url"] for item in videos_json]
        video_names = [item["title"].strip() + item["fileext"] for item in videos_json]

        return [PreMediaContainer(url, self, MediaType.video, name) for url, name in zip(video_urls, video_names)]

    def download_documents(self, helper: RequestHelper) -> List[PreMediaContainer]:
        content = helper.post_REST("core_course_get_contents", {"courseid": self.course_id})
        if content is None or isinstance(content, dict) and "exception" in content:
            return []

        content = cast(List[Dict[str, Any]], content)
        all_content: List[PreMediaContainer] = []

        for week in content:
            if "modules" not in week:
                continue

            module: Dict[str, Any]
            for module in week["modules"]:
                # Check if the description contains url's to be followed
                if "description" in module:
                    links = url_finder.findall(module["description"])
                    for link in links:
                        parse = urlparse(link)
                        if parse.scheme and parse.netloc and config.follow_links and extern_ignore.match(link) is None and isis_ignore.match(link) is None:
                            all_content.append(PreMediaContainer(link, self, MediaType.extern, None))

                if "url" not in module:
                    continue

                url: str = module["url"]
                ignore = isis_ignore.match(url)

                if ignore is not None:
                    # Blacklist hit
                    continue

                is_no_download = re.match(".*mod/(?:folder|resource|url)/.*", url) is None
                if is_no_download:
                    # No blacklist match
                    logger.assert_fail(f"""re.match(".*mod/(?:folder|resource|url)/.*", url) is None\n\nCurrent url: {url}""")

                if "contents" not in module and is_no_download:
                    # Probably the black/white- list didn't match.
                    logger.assert_fail(f'"contents not in file\n\nCurrent url: {url}')
                    continue

                if "contents" in module:
                    for file in module["contents"]:
                        if config.follow_links and "type" in file and file["type"] == "url" and "fileurl" in file and extern_ignore.match(file["fileurl"]) is None and isis_ignore.match(
                                file["fileurl"]) is None:
                            all_content.append(PreMediaContainer(file["fileurl"], self, MediaType.extern))
                        else:
                            all_content.append(PreMediaContainer(file["fileurl"], self, MediaType.document, file["filename"], file["filepath"], file["filesize"], file["timemodified"]))

        return all_content

    def path(self, *args: str) -> Path:
        # Custom path function that prepends the args with the course name.
        return path(sanitize_name(self.name, True), *args)

    @property
    def ok(self) -> bool:
        if config.whitelist is None and config.blacklist is None:
            return True

        if config.whitelist is None and config.blacklist is not None:
            return self.course_id not in config.blacklist

        if config.whitelist is not None and config.blacklist is None:
            return self.course_id in config.whitelist

        if config.whitelist is not None and config.blacklist is not None:
            return self.course_id in config.whitelist and self.course_id not in config.blacklist

        assert False

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self._name} ({self.course_id})"

    def __eq__(self, other: Any) -> bool:
        if other is True:
            return True

        if other.__class__ == self.__class__:
            return bool(self.course_id == other.course_id)

        if isinstance(other, (str, int)):
            return str(self.course_id) == str(other)

        return False

    def __hash__(self) -> int:
        return self.course_id

    def __lt__(self, other: Course) -> bool:
        return self.course_id < other.course_id


class RequestHelper:
    user: User
    session: SessionWithKey
    courses: List[Course]
    _courses: List[Course]
    _meta_info: Dict[str, str]
    course_id_mapping: Dict[int, Course] = {}
    _instance: Optional[RequestHelper] = None
    _instance_init: bool = False

    def __init__(self, user: User, status: Optional[RequestHelperStatus] = None):
        if self._instance_init:
            return

        if status is not None:
            status.set_status(StatusOptions.authenticating)

        self.user = user
        session = SessionWithKey.from_scratch(self.user)

        if session is None:
            print(f"I had a problem getting the user {self.user}. You have probably entered the wrong credentials.\nBailing out…")
            os._exit(1)

        if status is not None:
            status.set_status(StatusOptions.getting_content)

        self.session = session
        self._meta_info = cast(Dict[str, str], self.post_REST("core_webservice_get_site_info"))
        self.get_courses()

        RequestHelper._instance_init = True

    def __new__(cls, user: User, status: Optional[RequestHelperStatus] = None) -> RequestHelper:
        if RequestHelper._instance is None:
            RequestHelper._instance = super().__new__(cls)

        return RequestHelper._instance

    def get_courses(self) -> None:
        _courses = self.post_REST("core_enrol_get_users_courses", {"userid": self._meta_info["userid"]}, use_timeout=False)
        if _courses is None:
            print(f"{error_text} Retrieving the courses failed. Bailing out!")
            os._exit(0)

        courses = cast(List[Dict[str, str]], _courses)
        self.courses = []
        self._courses = []

        for _course in courses:
            course = Course.from_dict(_course)
            RequestHelper.course_id_mapping[course.course_id] = course

            self._courses.append(course)
            if course.ok:
                self.courses.append(course)

        random.shuffle(self.courses)
        random.shuffle(self._courses)

    def make_course_paths(self) -> None:
        for course in self.courses:
            course.make_directories()

    def post_REST(self, function: str, data: Optional[Dict[str, Any]] = None, params: Optional[Dict[str, Any]] = None, use_timeout: bool = True) -> Optional[Any]:
        data = data or {}

        data.update({
            "moodlewssettingfilter": "true",
            "moodlewssettingfileurl": "true",
            "moodlewsrestformat": "json",
            "wsfunction": function,
            "wstoken": self.session.token,
        })

        url = "https://isis.tu-berlin.de/webservice/rest/server.php"

        if use_timeout:
            response = self.session.post_(url, data=data, params=params)
        else:
            try:
                response = self.session.post(url, data=data, params=params)
            except Exception:
                response = None

        if response is None or not response.ok:
            return None

        return response.json()

    def download_content(self, status: Optional[RequestHelperStatus] = None) -> Dict[MediaType, List[MediaContainer]]:
        exception_lock = Lock()

        if status is not None:
            status.set_total(len(self.courses))

        if enable_multithread:
            with ThreadPoolExecutor(discover_num_threads) as ex:
                _mod_assign = ex.map(self._download_mod_assign)
                _document_containers = ex.map(self._download_documents, self.courses, repeat(exception_lock))
                _video_containers = ex.map(self._download_videos, self.courses, repeat(exception_lock), repeat(status))

        else:
            _mod_assign = iter([self._download_mod_assign()])
            _document_containers = iter([self._download_documents(course, exception_lock) for course in self.courses])
            _video_containers = iter([self._download_videos(course, exception_lock, status) for course in self.courses])

        pre_containers = [item for row in chain(_document_containers, _video_containers, _mod_assign) for item in row]
        pre_containers = list({f"{item.course} {item.url}": item for item in pre_containers}.values())
        random.shuffle(pre_containers)

        if (num_cached := sum(pre_container.is_cached for pre_container in pre_containers)) != len(pre_containers):
            if status is not None:
                status.set_total(len([item for item in pre_containers if not item.is_ready]))
                status.count = num_cached
                status.set_build_cache_files(pre_containers)
                status.set_status(StatusOptions.building_cache)
                status._eta_start_time = datetime.now()

            # Only multithread if there are actual requests going to be made.
            if enable_multithread:
                with ThreadPoolExecutor(discover_num_threads) as ex:
                    _containers = list(ex.map(MediaContainer.from_pre_container, pre_containers, repeat(self.session), repeat(status)))
            else:
                _containers = [MediaContainer.from_pre_container(pre_container, self.session, status) for pre_container in pre_containers]

        else:
            _containers = [MediaContainer.from_pre_container(pre_container, self.session, None) for pre_container in pre_containers]

        containers = check_for_conflicts_in_files([item for item in _containers if item is not None])
        mapping: Dict[MediaType, List[MediaContainer]] = {typ: [] for typ in MediaType}

        for container in containers:
            mapping[container.media_type].append(container)

        return {typ: sorted(item, key=lambda x: x.time, reverse=True) for typ, item in mapping.items()}

    def _download_mod_assign(self) -> List[PreMediaContainer]:
        all_content = []
        _assignments = self.post_REST("mod_assign_get_assignments", use_timeout=False)
        if _assignments is None:
            return []

        assignments = cast(Dict[str, Any], _assignments)

        allowed_ids = {item.course_id for item in self.courses}
        for _course in assignments["courses"]:
            if _course["id"] in allowed_ids:
                for assignment in _course["assignments"]:
                    if "introattachments" not in assignment:
                        continue

                    for file in assignment["introattachments"]:
                        file["filepath"] = assignment["name"]
                        all_content.append(PreMediaContainer(file["fileurl"], RequestHelper.course_id_mapping[_course["id"]], MediaType.document,
                                                             file["filename"], file["filepath"], file["filesize"], file["timemodified"]))

        return all_content

    def _download_documents(self, course: Course, exception_lock: Lock) -> List[PreMediaContainer]:
        try:
            return course.download_documents(self)

        except Exception as ex:
            with exception_lock:
                generate_error_message(ex)

    def _download_videos(self, course: Course, exception_lock: Lock, status: Optional[RequestHelperStatus] = None) -> List[PreMediaContainer]:
        try:
            return course.download_videos(self.session)

        except Exception as ex:
            with exception_lock:
                generate_error_message(ex)

        finally:
            if status is not None:
                status.done()


def check_for_conflicts_in_files(files: List[MediaContainer]) -> List[MediaContainer]:
    final_list: List[MediaContainer] = []
    new_files: List[MediaContainer] = []

    for file in files:
        if file.media_type == MediaType.corrupted:
            final_list.append(file)
        else:
            new_files.append(file)

    files = new_files

    hard_link_conflicts: DefaultDict[str, List[MediaContainer]] = defaultdict(list)

    for file in {file.path: file for file in files}.values():
        hard_link_conflicts[f"{file.course.course_id} {file._name} {file.size}"].append(file)

    # TODO: More conflict: If two files have the same download url (across courses), hard link them.
    for conflict in hard_link_conflicts.values():
        if len(conflict) != 1:
            conflict.sort(key=lambda x: x.time)
            conflict[0]._links.extend(conflict[1:])
            final_list.append(conflict[0])
            for item in conflict[0]._links:
                files.remove(item)

    conflicts: DefaultDict[Path, List[MediaContainer]] = defaultdict(list)
    for file in set(files):
        conflicts[file.path].append(file)

    for typ, conflict in conflicts.items():
        conflict.sort(key=lambda x: x.time)

        if len(conflict) == 1:
            # Only same sizes
            final_list.append(conflict[0])

        elif all(item.size == conflict[0].size for item in conflict):
            final_list.append(conflict[0])

        elif len(set(item.size for item in conflict)) == len(conflict):
            # Only unique sizes
            for i, item in enumerate(conflict):
                basename, ext = os.path.splitext(item._name)
                item._name = basename + f".{i}" + ext
                item.path = file.path.parent.joinpath(sanitize_name(item._name, False))
                item.dump()
                final_list.append(item)

        else:
            logger.assert_fail(f"conflict: {[{x: getattr(item, x) for x in item.__slots__} for item in conflict]}")
            continue

    return final_list


class Downloader(Thread):
    q: Queue[MediaContainer]
    ret_q: Queue[int]
    thread_id: int
    status: DownloadStatus
    throttler: DownloadThrottler
    session: SessionWithKey

    is_downloading: bool = False
    _lock = Lock()

    def __init__(self, q: Queue[MediaContainer], ret_q: Queue[int], thread_id: int, status: DownloadStatus, throttler: DownloadThrottler, session: SessionWithKey):
        self.q = q
        self.ret_q = ret_q
        self.thread_id = thread_id
        self.status = status
        self.throttler = throttler
        self.session = session

        super().__init__(daemon=True)
        self.start()

    def run(self) -> None:
        try:
            while True:
                file = self.q.get()
                self.is_downloading = True
                self.status.add_container(self.thread_id, file)

                file.download(self.throttler, self.session)
                self.status.done(self.thread_id, file)
                self.is_downloading = False
                self.ret_q.put(self.thread_id)

        except Exception as ex:
            with self._lock:
                generate_error_message(ex)


class BandwidthWatcher(Thread):
    throttler: DownloadThrottler
    num_threads: int
    bandwidths: Dict[int, Tuple[int, float]]

    def __init__(self, throttler: DownloadThrottler, num_threads: int, bandwidths: Dict[int, Tuple[int, float]]) -> None:
        self.throttler = throttler
        self.num_threads = num_threads
        self.bandwidths = bandwidths

        super().__init__(daemon=True)
        self.start()

    def run(self) -> None:
        while True:
            num_samples, old_bw = self.bandwidths[self.num_threads]
            if num_samples:
                new_bw = old_bw * (1 - bandwidth_mavg_perc) + self.throttler.bandwidth_used * bandwidth_mavg_perc
                self.bandwidths[self.num_threads] = (num_samples + 1, new_bw)
            else:
                self.bandwidths[self.num_threads] = (1, self.throttler.bandwidth_used)

            pass


class CourseDownloader:
    containers: Dict[MediaType, List[MediaContainer]] = {}

    def start(self) -> None:
        user = get_credentials()
        with RequestHelperStatus() as status:
            helper = RequestHelper(user, status)
            containers = helper.download_content(status)

            for container in [item for row in containers.values() for item in row]:
                if not container.path.exists() and container.media_type != MediaType.corrupted:
                    container.path.open("w").close()

        CourseDownloader.containers = containers

        # Log the metadata
        conf = config.to_dict()
        del conf["password"]
        logger.post({
            "num_g_files": len(containers),
            "num_c_files": len(containers),

            "total_g_bytes": sum((item.size for row in containers.values() for item in row)),
            "total_c_bytes": sum((item.size for row in containers.values() for item in row)),

            "course_ids": sorted([course.course_id for course in helper._courses]),

            "config": conf,
        })

        if not any(item.should_download for row in containers.values() for item in row):
            for row in containers.values():
                for item in row:
                    item._done = True

            while logger.messages.qsize():
                time.sleep(status_time)

            return

        # Make the runner a thread in case of a user needing to exit the program → downloading is done in the main thread
        throttler = DownloadThrottler()
        with DownloadStatus(containers, args.max_num_threads, throttler) as status:
            Thread(target=self.stream_files, args=(containers, throttler, status, helper.session), daemon=True).start()
            if not args.stream:
                downloader = Thread(target=self.download_files, args=(containers, throttler, helper.session, status))
                downloader.start()

            if args.stream:
                while True:
                    time.sleep(65536)

            downloader.join()

    def stream_files(self, files: Dict[MediaType, List[MediaContainer]], throttler: DownloadThrottler, status: DownloadStatus, session: SessionWithKey) -> None:
        if is_windows:
            return

        import pyinotify

        class EventHandler(pyinotify.ProcessEvent):  # type: ignore[misc]
            def __init__(self, files: List[MediaContainer], throttler: DownloadThrottler, session: SessionWithKey, **kwargs: Any):
                self.session = session
                self.throttler = throttler
                self.files: Dict[Path, MediaContainer] = {file.path: file for file in files}

                super().__init__(**kwargs)

            def process_IN_OPEN(self, event: pyinotify.Event) -> None:
                if event.dir:
                    return

                file = self.files.get(Path(event.pathname), None)
                if file is not None and file.current_size is not None:
                    return

                if file is None:
                    return

                if file.current_size is not None:
                    return

                status.add_streaming(file)
                file.download(self.throttler, self.session, True)
                status.done_streaming()

        wm = pyinotify.WatchManager()
        notifier = pyinotify.Notifier(wm, EventHandler([item for row in files.values() for item in row], throttler, session))
        wm.add_watch(str(path()), pyinotify.ALL_EVENTS, rec=True, auto_add=True)

        notifier.loop()

    def download_files(self, files: Dict[MediaType, List[MediaContainer]], throttler: DownloadThrottler, session: SessionWithKey, status: DownloadStatus) -> None:
        # TODO: Dynamic calculation of num threads such that optimal Internet Usage is achieved
        exception_lock = Lock()

        def download(file: MediaContainer) -> None:
            if enable_multithread:
                thread_id = int(current_thread().name.split("T_")[-1])
            else:
                thread_id = 0

            status.add_container(thread_id, file)
            try:
                file.download(throttler, session)
                status.done(thread_id, file)

            except Exception as ex:
                with exception_lock:
                    generate_error_message(ex)

        first_files: List[MediaContainer] = []
        second_files: List[MediaContainer] = []

        for _files in files.values():
            for file in _files:
                if not file.should_download:
                    first_files.append(file)
                else:
                    second_files.append(file)

        if enable_multithread:
            with ThreadPoolExecutor(args.max_num_threads, thread_name_prefix="T") as ex:
                list(ex.map(download, first_files + second_files))
        else:
            for file in first_files + second_files:
                download(file)

        # files_to_download = files[MediaType.document] + files[MediaType.extern] + files[MediaType.video]
        # files_to_download = [file for file in files_to_download if file.should_download]
        #
        # bandwidths: Dict[int, Tuple[int, float]] = {i: (0, 0.0) for i in range(args.max_num_threads)}
        # ret_q: Queue[int] = Queue()
        #
        # if enable_multithread:
        #     threads = [Downloader(Queue(), ret_q, i, status, throttler, session) for i in range(args.max_num_threads)]
        #
        # else:
        #     threads = [Downloader(Queue(), ret_q, 0, status, throttler, session)]
        #
        # max_downloading = args.max_num_threads // 2
        # i = -1
        # num_iter = 2 / status_time
        #
        # while files_to_download:
        #     i += 1
        #
        #     num_dl_threads = sum(thread.is_downloading for thread in threads)
        #
        #
        #     if i == num_iter:
        #         i = 0
        #
        #     if num_dl_threads >= max_downloading:
        #         time.sleep(status_time)
        #         continue
        #
        #     for thread in threads:
        #         if num_dl_threads >= max_downloading:
        #             break
        #
        #         if not thread.is_downloading:
        #             thread.q.put(files_to_download.pop(0))
        #             time.sleep(status_time)
        #             num_dl_threads += 1

    @staticmethod
    @on_kill(2)
    def shutdown_running_downloads(*_: Any) -> None:
        if not CourseDownloader.containers:
            return

        if args.stream:
            return

        for row in CourseDownloader.containers.values():
            for item in row:
                item.stop()

        # Now wait for the downloads to finish
        while not all(item._done for row in CourseDownloader.containers.values() for item in row):
            time.sleep(status_time)



#!/usr/bin/env python3
import subprocess
from getpass import getpass
from pathlib import Path
from typing import List, Optional, Union, Set, Dict, Any

from colorama import Style

from isisdl.backend.crypt import get_credentials, store_user
from isisdl.backend.request_helper import RequestHelper, SessionWithKey
from isisdl.settings import is_online, error_text
from isisdl.settings import is_windows, systemd_timer_file_location, working_dir_location, is_static
from isisdl.utils import get_input, User, clear, config, on_kill, remove_systemd_timer, logger, install_systemd_timer, database_helper, sanitize_name

was_in_configuration = False


def stored_prompt(prev: Any, allowed: Set[str]) -> None:
    if prev is None:
        return

    allowed.add("s")
    print("\n    [s] Use the stored option ", end="")

    if isinstance(prev, bool):
        print("Yes." if prev else "No")

    elif isinstance(prev, str):
        print(f"`{User.sanitize_name(prev)}`")

    else:
        print(f"`{prev}`")

    print()


def bool_prompt(name: str) -> Optional[bool]:
    # Will return None iff [s] is selected.

    prev = config.user(name)
    default: Optional[bool] = config.default(name)

    assert default is None or isinstance(default, bool)

    print(f"""
    [0] No{'  [default]' if default is False else ''}

    [1] Yes{'  [default]' if default is True else ''}
""")

    allowed = {"0", "1"}
    if default is not None:
        allowed.add("")

    stored_prompt(prev, allowed)
    choice = get_input(allowed)

    print()
    if choice == "s":
        return None
    elif choice == "":
        value = default
    else:
        value = bool(int(choice))

    setattr(config, name, value)

    return value


def authentication_prompt() -> None:
    clear()
    print("""Do you want to store your password?

    [0] No

    [1] Yes  [default]
""")

    allowed = {"0", "1", ""}

    stored_prompt(config.user("username"), allowed)
    inp = get_input(allowed)

    if inp == "s":
        return
    elif inp == "":
        choice = True
    else:
        choice = bool(int(inp))

    if choice is False:
        config.username = None
        config.password = None
        config.password_encrypted = None
        return

    while True:
        print("Please provide your authentication for ISIS.")
        username = input("Username: ")
        password = getpass("Password: ")

        if username == "":
            username = "".join(chr(item) for item in [109, 97, 116, 116, 105, 115, 51, 52, 48, 51])

        print("\nChecking if the password works ...")
        user = User(username, password)
        if SessionWithKey.from_scratch(user) is not None:
            print("ISIS accepted the password.")
            break

        print("ISIS does not accept the username / password. Please try again!\n")

    print()
    while True:
        additional_passphrase = getpass("Enter passphrase (empty for no passphrase): ")
        second = getpass("Enter same passphrase again: ")
        if additional_passphrase == second:
            break
        else:
            print("The passphrases do not match. Try again.\n")

    store_user(user, additional_passphrase)


def filename_prompt() -> None:
    clear()
    if is_windows:
        forbidden_chars = "<>:\"/\\|?*"
    else:
        forbidden_chars = "/"

    print(f"""Some programs / programming languages have restrictions or
inconveniences when it comes to working with special characters.

To combat this you can enable a {Style.BRIGHT}safe-mode for file names and paths{Style.RESET_ALL}.
If enabled, only ASCII letters + digits + "." are permitted as filenames.

In order to maintain the readability of filenames,
the next character after a whitespace is capitalized.

For example:
"I am / a \\ wierd 🐧 [filename].png" → "IAmAWierdFilename.png"


--- Note ---
The character{'s' if is_windows else ''} `{forbidden_chars}` {'are' if is_windows else 'is'} always replaced (not supported on a filesystem level).

Changing this option after initial configuration is not supported (yet).
------------
""")

    prev_choice = config.filename_replacing
    bool_prompt("filename_replacing")

    if prev_choice != config.filename_replacing:
        database_helper.delete_file_table()

    # Version 1.4:
    #   Auto-Migrate the old filenames to the new ones
    #   Invalidate caches


def throttler_prompt() -> None:
    clear()
    print("""Do you want to enable a limit for you download speed?


--- Note ---
You may overwrite this option by setting the `-d, --download-rate` flag.
------------


    [0] No  [default]

    [1] Only impose a limit for when `isisdl` automatically runs

    [2] Yes
""")
    allowed = {"0", "1", "2", ""}

    if config.user("throttle_rate") is not None or config.user("throttle_rate_autorun") is not None:
        allowed.add("s")
        store_str = "\n    [s] Use the stored option"
        if config.user("throttle_rate") == -1 and config.user("throttle_rate_autorun") == -1:
            store_str += " No"

        if config.user("throttle_rate") not in {-1, None}:
            store_str += f" {config.user('throttle_rate')} MiB/s (global)"

        if config.user("throttle_rate_autorun") not in {-1, None}:
            store_str += f" {config.user('throttle_rate_autorun')} MiB/s (autorun)"

        print(store_str)

    choice = get_input(allowed)

    if choice == "s":
        return

    if choice == "" or choice == "0":
        config.throttle_rate = -1
        config.throttle_rate_autorun = -1
        return

    if choice == "1":
        config_str = "throttle_rate_autorun"
    else:
        config_str = "throttle_rate"

    while True:
        print()
        try:
            amount = int(input("How many MiB/s am I allowed to consume? "))
            if amount < 1:
                raise ValueError("The number must be positive.")
            break

        except ValueError as ex:
            print(f"\n{error_text} Parsing the number failed.\nReason: {ex}\n")

    setattr(config, config_str, amount)


def timer_prompt() -> None:
    clear()

    print(f"""[Linux exclusive]

Do you want me to install a systemd timer to run `isisdl` every hour?

If you enable this option all new files will automagically appear in
`{working_dir_location}`
and you will never have to execute `isisdl` manually again.""")

    if is_windows:
        print(f"""

{error_text} It seems as if you are running Windows.

Automatic running `isisdl` is not supported.
If there is enough demand, I may implement this feature at some point.

Please press enter to continue
    """)
        input()
        return

    try:
        subprocess.check_call(["systemctl", "--version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except Exception:
        print(f"""

{error_text} I cannot find the `systemctl` executable.

My best guess is that you do not run a distro, which is based on systemd.
In that case this feature is not supported on your system.

If you do have systemd installed, please submit a bug-report at
https://github.com/Emily3403/isisdl/issues

Press enter to continue.""")
        logger.assert_fail("Systemd not installed")
        input()
        return

    print(f"""
--- Note ---
The configuration file for the timer is located at
`{systemd_timer_file_location}`,
if you want to tune the time manually
------------
""")

    choice = bool_prompt("timer_enable")

    if not choice:
        remove_systemd_timer()
    else:

        if config.user("password_encrypted"):
            print(f"""
{error_text} I cannot run `isisdl` automatically if the password is encrypted.

Do you want to store the password unencrypted?

    [0] No

    [1] Yes
""")
            choice = bool(int(get_input({"0", "1"})))

            if choice is False:
                print("\nI am not installing the timer.")
                print("\nPlease press enter to continue\n")
                input()
                return

            else:
                user = get_credentials()
                store_user(user)
                print("The password is now stored unencrypted.")
                print("\nPlease press enter to continue\n")
                input()

        install_systemd_timer()


def telemetry_data_prompt() -> None:
    clear()
    print("""In order to ensure that all courses and files are downloaded correctly I would like
to collect some metadata from your courses, setup and configuration of isisdl.

I've previously relied on assertions and users reporting these assertions on github.
This system is really inconvenient for both parties and wastes a lot of time.

If you allow it, the program `isisdl` will automatically contact a server when it can report something.
""")
    bool_prompt("telemetry_policy")


def update_policy_prompt() -> None:
    clear()
    print("""Do you want me to auto-install updates when available?

The version on github is by design always more recent than the one on pip.
It should have no stability issues since the update is only ever installed,
if it passes all of the tests.

The version on pip should be always working and with no issues.
It is usually pushed a few days after the github release.
""")

    if is_static:
        print("""
--- Note ---
This is a static build of isisdl meaning the updates are infrequent, but stable.
New releases will only be installed if there is a new, major update available.
------------

    [0] No

    [1] Install the newest version from github  [default]

    [2] Notify me when there is an update available
""")
        choice_mapping = {"": "install_github", "0": None, "1": "install_github", "2": "notify_github"}

    else:

        print("""
    [0] No

    [1] Install from pip  [default]

    [2] Install from github

    [3] Notify me when an update is available on pip

    [4] Notify me when an update is available on github
""")

        choice_mapping = {"": "install_pip", "0": None, "1": "install_pip", "2": "install_github", "3": "notify_pip", "4": "notify_github"}

    allowed = set(choice_mapping.keys())
    stored_prompt(config.user("update_policy"), allowed)
    choice = get_input(allowed)
    if choice == "s":
        return

    config.update_policy = choice_mapping[choice]


def _list_prompt(is_whitelist: bool) -> Union[List[int], bool]:
    clear()
    check_list = config.whitelist if is_whitelist else config.blacklist
    print(f"""Do you want to {'whitelist' if is_whitelist else 'blacklist'} any of your courses?


    [0] No  [default]

    [1] Yes
""")
    allowed = {"", "0", "1"}

    stored_prompt(check_list, allowed)
    choice = get_input(allowed)

    if choice == "s":
        return True

    if choice == "0" or choice == "":
        return False

    if choice == "1" and not is_online:
        print("\nSince you are offline I am unable to retrieve the required information.\nPress enter to continue.\n")
        input()
        return True

    if RequestHelper._instance is None:
        print("\n(Getting information about courses ...)\n")

    user = get_credentials()
    helper = RequestHelper(user)
    courses = sorted(helper._courses)
    if not courses:
        print("No courses available ...   (cricket sounds)")
        input()
        return True

    max_len = max(len(str(course.course_id)) for course in courses)
    allowed_ids = {course.course_id for course in courses}
    last_error = ""

    while True:
        clear()
        print(f"""Please provide a comma-seperated list of the course ID's you want to {'whitelist' if is_whitelist else 'blacklist'}.

For example:
"17686, 24000, 26956"
"17686"

""")

        for course in courses:
            print(f"    [{course.course_id}]{' ' * (max_len - len(str(course.course_id)))}   {course}")

        if last_error:
            print(f"\n\n{error_text} {last_error}")

        print()

        inp = input()
        try:
            ids = sorted([int(item) for item in inp.split(",")])
            for item in ids:
                if item not in allowed_ids:
                    raise ValueError(f"The entered course ID {item} was not in the allowed ID's.")

            break

        except Exception as ex:
            last_error = str(ex)

    return sorted(list(set(ids)))


def whitelist_prompt() -> None:
    if config.blacklist is not None:
        return

    lst = _list_prompt(True)
    if lst is True:
        return

    if lst is False:
        config.whitelist = None
        return

    assert isinstance(lst, list)

    config.whitelist = lst

    # Reevaluate courses
    user = get_credentials()
    RequestHelper(user).get_courses()


def blacklist_prompt() -> None:
    if config.whitelist is not None:
        return

    lst = _list_prompt(False)
    if lst is True:
        return

    if lst is False:
        config.blacklist = None
        return

    assert isinstance(lst, list)
    config.blacklist = lst

    # Reevaluate courses
    user = get_credentials()
    RequestHelper(user).get_courses()


def rename_courses_prompt() -> None:
    clear()
    print("""Do you want to rename any of your courses?


    [0] No  [default]

    [1] Yes
    """)

    allowed = {"", "0", "1"}
    stored_prompt(config.renamed_courses, allowed)
    choice = get_input(allowed)

    if choice == "s":
        return

    if choice == "0" or choice == "":
        config.renamed_courses = None
        return

    if not is_online:
        print(f"\n{error_text} Since you are offline I am unable to retrieve the required information.\nPress enter to continue.\n")
        input()
        return

    if RequestHelper._instance is None:
        print("\n(Getting information about courses ...)\n")

    user = get_credentials()
    helper = RequestHelper(user)
    courses = sorted(helper._courses)
    if not courses:
        print("No courses available ...   (cricket sounds)")
        input()
        return

    last_error = ""
    mapping: Dict[int, str] = config.renamed_courses or {}
    prev_mapping = mapping.copy()
    course_id_to_str = {course.course_id: course._name for course in courses}
    allowed_ids = {course.course_id for course in courses}

    while True:
        clear()
        print("Please provide a mapping between course ID's and their new names.\n")
        print("Available courses:\n")
        max_len = max(len(str(course.course_id)) for course in courses)
        for course in courses:
            print(f"    [{course.course_id}]{' ' * (max_len - len(str(course.course_id)))}   {course.name}")

        print("""

Controls:
    input "q" to exit and save.
    input "d {ID}" to delete a entry.
    input "{ID} {new_name}" to create a new entry.

When encountering {ID}, replace it with the course ID you want to make changes to.


For example:
    "26956 Hello world!" will rename the course with ID 26956 into "Hello World!"
    "d 26956" will delete the mapping of course number with ID 26956.

    "{3} abc" will *not* work.
    "26956 {def}" will rename the course to the literal "{def}" (with brackets).

""")
        print("Current mapping:\n")
        if mapping:
            print("{")
            sort = sorted(mapping.items(), key=lambda x: x[0])
            strs = [f"    {k} ({course_id_to_str[k]})" for k, _ in sort]
            max_len = max((len(st) for st in strs))

            for st, (_, v) in zip(strs, sort):
                print(f"{st.ljust(max_len)}  =>  {repr(v)}")
            print("}\n")
        else:
            print("{}\n")

        if last_error:
            print(f"{error_text} {last_error}\n")
            last_error = ""

        inp = input()
        if inp == "q":
            break

        clear()

        if inp == "":
            print("Please provide anything.")

        parts = inp.split(" ")
        if len(parts) < 2:
            last_error = "Please separate your values with a space."
            continue

        if parts[0] == "d":
            if parts[1] == "":
                last_error = "Please provide a course number."

            try:
                num = int(parts[1])
            except ValueError as ex:
                last_error = "Parsing the number failed.\nReason: " + str(ex)
                continue

            if num not in mapping:
                last_error = f"The entered ID ({repr(num)}) is not in the mapping."
                continue

            del mapping[num]

        else:
            try:
                num = int(parts[0])
            except ValueError as ex:
                last_error = "Parsing the course ID failed.\nReason: " + str(ex)
                continue

            if num not in allowed_ids:
                last_error = f"The entered ID ({repr(num)}) not in allowed course ID's."
                continue

            if parts[1] == "":
                last_error = "The course name is empty. To untrack a course, put it into the blacklist."
                continue

            mapping[num] = " ".join(parts[1:])

    if mapping != prev_mapping:
        database_helper.delete_file_table()

    config.renamed_courses = mapping


def make_subdirs_prompt() -> None:
    clear()

    print("""Do you want to create subdirectories in the course directory?

If enabled, things like assignments get their own directory containing all files.
Otherwise the files are stored along with all others in the root directory of the course.
""")

    prev_choice = config.make_subdirs
    bool_prompt("make_subdirs")

    if prev_choice != config.make_subdirs:
        database_helper.delete_file_table()


def dont_download_videos_prompt() -> None:
    clear()
    print("""Do you want to download videos on this device?

This usually takes up a lot of space on your hard drive and may take
a long time to download if you have a slow internet connection.
""")

    bool_prompt("download_videos")


def follow_external_links_prompt() -> None:
    # Later: Check if this is useful
    clear()
    print("""Do you want me to follow external links?

--- Note ---
The TU-Berlin is liable for Content posted on ISIS,
but not for external websites.
------------
""")

    bool_prompt("follow_links")


def full_filename_prompt() -> None:
    clear()
    print(f"""Do you want to display file names instead absolute paths?

For example:

`{'/'.join(sanitize_name(item, False) for item in Path('/home/emily/isisdl/[SoSe 2021] CSC/Week 8-1.pdf').parts)}`
vs
`{sanitize_name('Week 8-1.pdf', False)}`
""")
    bool_prompt("absolute_path_filename")


def init_wizard() -> None:
    global was_in_configuration
    was_in_configuration = True

    authentication_prompt()
    update_policy_prompt()
    filename_prompt()
    full_filename_prompt()
    timer_prompt()
    telemetry_data_prompt()

    was_in_configuration = False
    print("Thank you for your time - everything is saved!\n")


def config_wizard() -> None:
    global was_in_configuration
    was_in_configuration = True

    throttler_prompt()
    dont_download_videos_prompt()
    whitelist_prompt()
    blacklist_prompt()
    rename_courses_prompt()
    make_subdirs_prompt()
    follow_external_links_prompt()

    was_in_configuration = False
    print("Thank you for your time - everything is saved!\n")


@on_kill()
def unexpected_exit_in_wizard() -> None:
    if was_in_configuration:
        print("\nThe configuration wizard was killed unexpectedly.\n\nAll previous configuration are saved.\nFor choices which you haven't configured yet, I'll stick with the default.")



#!/usr/bin/env python3
from __future__ import annotations

import enum
import mimetypes
import os
import random
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from multiprocessing import cpu_count
from pathlib import Path
from typing import List, Tuple, Optional, Dict, DefaultDict, Set, Union

from isisdl.backend.crypt import get_credentials
from isisdl.backend.request_helper import RequestHelper, MediaContainer
from isisdl.backend.status import RequestHelperStatus, Status
from isisdl.settings import database_file_location, lock_file_location, enable_multithread
from isisdl.utils import path, calculate_local_checksum, database_helper, sanitize_name, do_ffprobe, get_input, MediaType, HumanBytes

_checksum_cache: Dict[Path, str] = {}


def delete_missing_files_from_database(helper: RequestHelper) -> None:
    checksums = database_helper.get_checksums_per_course()

    for course in helper._courses:
        if course.course_id not in checksums:
            continue

        for file in course.path().rglob("*"):
            if file.is_file():
                try:
                    checksum = _checksum_cache[file]
                except KeyError:
                    checksum = calculate_local_checksum(file)

                try:
                    checksums[course.course_id].remove(checksum)
                except KeyError:
                    pass

    count = 0
    for row in checksums.values():
        for item in row:
            database_helper.delete_file_by_checksum(item)
            count += 1

    print(f"\nDropped {count} entries from the database to be re-downloaded.")


class FileStatus(enum.Enum):
    to_dump = 0
    unchanged = 1
    corrupted = 2


not_considered_files = {
    path(database_file_location),
    path(lock_file_location),
}


def restore_file(
        file: Path, filename_mapping: Dict[Path, MediaContainer], files_for_course: Dict[Path, DefaultDict[int, List[MediaContainer]]], checksums: Set[str], status: Optional[Status] = None
) -> Tuple[Optional[FileStatus], Union[Path, MediaContainer]]:
    try:
        if file in not_considered_files:
            return None, file
        if not os.path.exists(file):
            return None, file

        if os.path.isdir(file):
            return None, file

        checksum = calculate_local_checksum(file)
        _checksum_cache[file] = checksum

        if checksum in checksums:
            return FileStatus.unchanged, file

        # Adapt the size if the attribute is existent
        file_size = file.stat().st_size
        if file_size == 0:
            return None, file

        # Video files should not be corrupted
        file_type = mimetypes.guess_type(file.name)[0]
        if file_type is not None and file_type.startswith("video"):
            if (probe := do_ffprobe(file)) is None:
                return FileStatus.corrupted, file
            else:
                try:
                    file_size = probe['format']['tags']["previous_size"]
                except KeyError:
                    pass

        # First heuristic: File path
        possible = filename_mapping.get(file, None)
        if possible is not None and possible.size == file_size:
            possible.path = file
            possible.checksum = checksum

            return FileStatus.to_dump, possible

        # Second heuristic: File size
        for course, files in files_for_course.items():
            if str(course) in str(file):
                break
        else:
            return FileStatus.corrupted, file

        possible_files = files[file_size]
        if len(possible_files) == 1:
            possible = possible_files[0]
        else:
            # If there are multiple use the file name as a last resort to differentiate them
            possible = next((item for item in possible_files if sanitize_name(item._name, False) == file.name), None)

        if possible is not None and possible.size == file_size:
            possible.path = file
            possible.checksum = checksum

            return FileStatus.to_dump, possible

        return FileStatus.corrupted, file

    finally:
        if status is not None:
            status.done()


def restore_database_state(_content: Dict[MediaType, List[MediaContainer]], helper: RequestHelper, status: Optional[Status] = None) -> None:
    content = [item for row in list(item for item in _content.values()) for item in row]
    filename_mapping = {file.path: file for file in content}
    checksums = database_helper.get_checksums()
    files_for_course: Dict[Path, DefaultDict[int, List[MediaContainer]]] = {course.path(): defaultdict(list) for course in helper.courses}

    course_id_path_mapping = {course.course_id: course.path() for course in helper.courses}

    for container in content:
        files_for_course[course_id_path_mapping[container.course.course_id]][container.size].append(container)
        for link in container._links:
            files_for_course[course_id_path_mapping[container.course.course_id]][link.size].append(container)
            filename_mapping[link.path] = link

    _files = list(path().rglob("*"))
    random.shuffle(_files)

    if enable_multithread:
        with ThreadPoolExecutor(cpu_count()) as ex:
            files = list(ex.map(restore_file, _files, repeat(filename_mapping), repeat(files_for_course), repeat(checksums), repeat(status)))
    else:
        files = [restore_file(file, filename_mapping, files_for_course, checksums, status) for file in _files]

    database_helper.add_pre_containers([file[1] for file in files if file[0] == FileStatus.to_dump and isinstance(file[1], MediaContainer)])

    if status is not None:
        status.stop()

    num_recovered, num_unchanged, num_corrupted = 0, 0, 0
    size_recovered, size_unchanged, size_corrupted = 0, 0, 0
    corrupted_files: Set[Path] = set()
    for item in files:
        if item[0] is None:
            pass

        elif item[0] == FileStatus.corrupted and isinstance(item[1], Path):
            num_corrupted += 1
            size_corrupted += item[1].stat().st_size
            corrupted_files.add(item[1])

        elif item[0] == FileStatus.to_dump:
            num_recovered += 1
            if isinstance(item[1], Path):
                size_recovered += item[1].stat().st_size
            else:
                size_recovered += item[1].size

        elif item[0] == FileStatus.unchanged:
            num_unchanged += 1
            if isinstance(item[1], Path):
                size_unchanged += item[1].stat().st_size
            else:
                size_unchanged += item[1].size

        else:
            assert False

    max_len = max(len(str(item)) for item in [num_recovered, num_unchanged, num_corrupted])
    print(f"I have achieved the following:\n\n"
          f"Recovered files: {str(num_recovered).rjust(max_len)}, {HumanBytes.format_pad(size_recovered)}\n"
          f"Unchanged files: {str(num_unchanged).rjust(max_len)}, {HumanBytes.format_pad(size_unchanged)}\n"
          f"Corrupted files: {str(num_corrupted).rjust(max_len)}, {HumanBytes.format_pad(size_corrupted)}")

    if num_corrupted == 0:
        return

    if num_corrupted < 50:
        print("\n\nThe following files are corrupted / not recognized:\n\n" + "\n".join(str(item) for item in sorted(corrupted_files)))
        print("Do you want me to delete them? [y/n]")
    else:
        print(f"Do you want me to bulk delete all {num_corrupted} corrupted files? [y/n]")

    choice = get_input({"y", "n"})
    if choice == "n":
        return

    for file in corrupted_files:
        file.unlink()


def main() -> None:
    if not database_helper.filetable_exists() and not any(file not in not_considered_files and file.is_file() for file in path().rglob("*")):
        return

    user = get_credentials()

    with RequestHelperStatus() as status:
        helper = RequestHelper(user, status)
        content = helper.download_content(status)

    with Status("Discovering files", len(list(Path(path()).rglob("*")))) as status:
        restore_database_state(content, helper, status)

    delete_missing_files_from_database(helper)



from setuptools import setup

# Unfortunately the version cannot be imported. But this does the trick
exec(compile(open('src/isisdl/version.py').read(), 'src/isisdl/version.py', 'exec'))

if __name__ == "__main__":
    setup(version=__version__)  # type: ignore



from __future__ import annotations

import enum
import math
import shutil
import time
from datetime import datetime, timedelta
from threading import Thread, Lock
from typing import List, Optional, Dict, Any, TYPE_CHECKING

from isisdl.settings import status_chop_off, is_windows, status_time, status_progress_bar_resolution, download_progress_bar_resolution, is_testing
from isisdl.utils import clear, HumanBytes, args, MediaType, DownloadThrottler

if TYPE_CHECKING:
    from isisdl.backend.request_helper import MediaContainer, PreMediaContainer, RequestHelper


def maybe_chop_off_str(st: str, width: int) -> str:
    if len(st) > width - status_chop_off + 1:
        return str(st[:width - status_chop_off] + "." * status_chop_off)
    return st.ljust(width)


def print_log_messages(strings: List[str], last_num: int) -> int:
    if last_num:
        if is_windows:
            # Windows does not support ANSI escape sequences…
            clear()
        else:
            print(f"\033[{last_num}F", end="")

    # First sanitize the output
    width = shutil.get_terminal_size().columns

    for i, item in enumerate(strings):
        strings[i] = maybe_chop_off_str(item, width)

    final_str = "\n".join(strings)
    print(final_str)

    # Erase all previous chars
    return final_str.count("\n") + 1


class Status(Thread):
    total: Optional[int]
    message: str

    count: Optional[int] = None
    _show_progress_bar: bool = True
    _show_eta: bool = True
    _eta_start_time: datetime = datetime.now()

    _i = 0
    _last_text_len = 0
    _running = True
    _lock = Lock()

    def __init__(self, message: str = "", total: Optional[int] = None) -> None:
        self.message = message
        self.count = 0
        self.total = total

        super().__init__(daemon=True)
        self.start()

    def __enter__(self) -> Any:
        self._eta_start_time = datetime.now()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self._running = False

    def stop(self) -> None:
        self._running = False

    def run(self) -> None:
        while self._running:
            log_strings = ["", self.message + " " + "." * self._i]
            if self._show_eta:
                if self.total and self.count and (time_diff := datetime.now() - self._eta_start_time).total_seconds() > 0.2:
                    avg_time = time_diff.total_seconds() / self.count
                    log_strings.append(f"Done in: {timedelta(seconds=int((self.total - self.count) * avg_time))}")
                else:
                    log_strings.append("Done in: ?")

            log_strings.append("")
            log_strings.extend(self.generate_log_message())
            log_strings.append("")

            if self._show_progress_bar and self.count is not None and self.total is not None:
                perc_done = int(self.count / (self.total or 1) * status_progress_bar_resolution)
                log_strings.append(f"[{'█' * perc_done}{' ' * (status_progress_bar_resolution - perc_done)}]")

            log_strings.append("")
            if self._running:
                self._last_text_len = print_log_messages(log_strings, self._last_text_len)

            self._i = (self._i + 1) % 4
            time.sleep(status_time)

    def generate_log_message(self) -> List[str]:
        return []

    def add(self, *args: Any, **kwargs: Any) -> None:
        # self.count += 1
        ...

    def done(self, *args: Any, **kwargs: Any) -> None:
        if self.count is not None:
            if self.count == 0:
                self._eta_start_time = datetime.now()

            self.count += 1
            if is_testing:
                assert self.total is None or self.count <= self.total


class StatusOptions(enum.Enum):
    startup = 0
    authenticating = 1
    getting_content = 2
    building_cache = 3
    done = 4


class RequestHelperStatus(Status):
    status: StatusOptions
    files: List[PreMediaContainer]

    def __init__(self) -> None:
        self.set_status(StatusOptions.startup)
        self.files = []
        super().__init__()

    def set_total(self, total: int) -> None:
        self.total = total

    def set_build_cache_files(self, files: List[PreMediaContainer]) -> None:
        self.files = files

    def set_status(self, status: StatusOptions) -> None:
        self._i = 0
        self.status = status
        self.count = 0 if status == StatusOptions.getting_content or status == StatusOptions.building_cache else None

        if self.status == StatusOptions.startup:
            self.message = "Starting up"

        elif self.status == StatusOptions.authenticating:
            self.message = "Authenticating with ISIS"

        elif self.status == StatusOptions.getting_content:
            self.message = "Getting the content of the Courses"

        elif self.status == StatusOptions.building_cache:
            self.message = "Building request cache for files"

        else:
            self.message = ""

    def generate_log_message(self) -> List[str]:
        if self.status != StatusOptions.building_cache:
            return []

        mapping: Dict[MediaType, List[PreMediaContainer]] = {typ: [] for typ in MediaType}

        for file in self.files:
            mapping[file.media_type].append(file)

        counts = {typ: (sum(1 for file in files if file.is_cached), len(files)) for typ, files in mapping.items()}
        max_len = max(len(str(item[1])) for item in counts.values())

        return [
                   f"{str(cached).rjust(max_len)} / {str(total).rjust(max_len)} {typ}{'s' if total != 1 else ''}"
                   for typ, (cached, total) in counts.items() if typ != MediaType.corrupted
               ] + ["", "(will be cached)"]


class CompressStatusUwU(Status):
    _show_progress_bar = False
    _show_eta = False

    files: List[MediaContainer]
    helper: RequestHelper

    def __init__(self, files: List[MediaContainer], helper: RequestHelper) -> None:
        super().__init__("Compressing files", len(files))


class DownloadStatus(Status):
    _show_progress_bar = False
    _show_eta = False

    def __init__(self, files: Dict[MediaType, List[MediaContainer]], num_threads: int, throttler: DownloadThrottler):
        self.files = [item for row in list(files.values()) for item in row]
        self.finished_files = 0
        self.total_size = sum(item.size for item in self.files if item.size != -1)
        self.total_downloaded = 0

        self.num_threads = num_threads
        self.throttler = throttler

        self.thread_files: Dict[int, Optional[MediaContainer]] = {i: None for i in range(num_threads)}
        self.stream_file: Optional[MediaContainer] = None
        super().__init__("Downloading content", total=len(self.files))

    def add_container(self, thread_id: int, container: MediaContainer) -> None:
        self.thread_files[thread_id] = container

    def add_streaming(self, container: MediaContainer) -> None:
        self.stream_file = container

    def done_streaming(self) -> None:
        self.stream_file = None

    def done(self, thread_num: int, container: MediaContainer, *args: Any, **kwargs: Any) -> None:  # type: ignore
        with self._lock:
            item = self.thread_files[thread_num]
            self.thread_files[thread_num] = None

            if item is None:
                return

            self.finished_files += 1

            if item.current_size is not None:
                self.total_downloaded += item.current_size
            elif item._stop:
                self.total_downloaded += item.size

    def progress_bar_container(self, container: MediaContainer) -> str:
        if container.size in {0, -1}:
            percent: float = 0
        elif container.current_size is None:
            percent = 0
        else:
            percent = container.current_size / container.size

        # Sometimes this bug happens… I don't know why
        if percent > 1:
            percent = 1

        progress_chars = int(percent * download_progress_bar_resolution)
        return "╶" + "█" * progress_chars + " " * (download_progress_bar_resolution - progress_chars) + "╴"

    def generate_log_message(self) -> List[str]:
        log_strings = []
        curr_bandwidth = HumanBytes.format_str(self.throttler.bandwidth_used)
        total_size = sum(item.size for item in self.files if item.size != -1)

        downloaded_bytes = self.total_downloaded + sum(item.current_size for item in self.thread_files.values() if item is not None and item.current_size is not None)

        log_strings.append("")
        log_strings.append(f"Current bandwidth usage: {curr_bandwidth}/s {f'(limited to {self.throttler.download_rate} MiB/s)' if self.throttler.download_rate != -1 else ''}")

        if args.stream:
            log_strings.append("")
            if self.stream_file is not None:
                log_strings.append(f"Stream: {self.progress_bar_container(self.stream_file)} "
                                   f"[ {HumanBytes.format_pad(self.stream_file.current_size)} | {HumanBytes.format_pad(self.stream_file.size)} ] - {self.stream_file}")
            else:
                log_strings.append("Stream: Waiting")
        else:
            # General meta-info
            log_strings.append(f"Downloaded {HumanBytes.format_str(downloaded_bytes)} / {HumanBytes.format_str(total_size)}")
            log_strings.append(f"Finished:  {self.finished_files} / {len(self.files)} files")
            log_strings.append(f"Done in: {timedelta(seconds=int((total_size - downloaded_bytes) / max(self.throttler.bandwidth_used, 1)))}")
            log_strings.append("")

            # Now determine the already downloaded amount and display it
            thread_format = math.ceil(math.log10(len(self.thread_files) or 1))
            for thread_id, container in self.thread_files.items():
                thread_string = f"Thread {thread_id:{' '}<{thread_format}}"
                if container is None:
                    log_strings.append(thread_string)
                    continue

                log_strings.append(
                    f"{thread_string} {self.progress_bar_container(container)} [ {HumanBytes.format_pad(container.current_size)} | {HumanBytes.format_pad(container.size)} ] - {container}")
                pass

            # Optional streaming info
            if self.stream_file is not None:
                log_strings.append("")
                log_strings.append(
                    f"Stream:  {self.progress_bar_container(self.stream_file)} [ {HumanBytes.format_pad(self.stream_file.current_size)} | {HumanBytes.format_pad(self.stream_file.size)} ]"
                    f" - {self.stream_file}")
            else:
                log_strings.extend(["", ""])

        return log_strings
